import json
from typing import (
    List,
    Iterable,
    Optional,
    Dict,
    Any,
    Callable,
    Tuple,
    Iterator,
    Union,
)

from pyspark.sql import DataFrame
from pyspark.sql.types import Row
from pyspark import SparkFiles
from requests import status_codes, Response

from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.register import register
from spark_pipeline_framework.utilities.api_helper.http_request import (
    HelixHttpRequest,
    RequestType,
)
from spark_pipeline_framework.utilities.oauth2_helpers.v2.oauth2_client_credentials_flow import (
    OAuth2ClientCredentialsFlow,
    OAuth2Credentails,
)

RESPONSE_PROCESSOR_TYPE = Callable[
    [Response, Any],
    Tuple[Any, bool],
]

REQUEST_GENERATOR_TYPE = Callable[
    [DataFrame, Optional[ProgressLogger]], Iterator[Tuple[HelixHttpRequest, Any]]
]


class HttpDataReceiverProcessor:
    @staticmethod
    def create_access_token(
        *, auth_url: str, credentials: Optional[OAuth2Credentails]
    ) -> str:
        """
        Factory function to create access token

        :param auth_url: OAuth token URL
        :param credentials: authentication credentails like client id/secrets
        """
        assert credentials
        oauth2_client_credentials_flow: OAuth2ClientCredentialsFlow = (
            OAuth2ClientCredentialsFlow(
                auth_url=auth_url,
                auth_credentails=credentials,
                progress_logger=None,
            )
        )

        access_token: Optional[str] = oauth2_client_credentials_flow.get_token()
        assert access_token
        print(f"Received token from {auth_url}: {access_token}")

        return access_token

    @staticmethod
    def process_rows(
        partition_index: int,
        rows: Iterable[Row],
        response_processor: RESPONSE_PROCESSOR_TYPE,
        raise_error: bool,
        credentials: Optional[OAuth2Credentails],
        auth_url: Optional[str],
        cert: Optional[Union[str, Tuple[str, str]]] = None,
        verify: Optional[Union[bool, str]] = None,
    ) -> List[Row]:
        register()

        result: List[Row] = []
        headers = {}

        if credentials or auth_url:
            assert credentials
            assert auth_url
            auth_token = HttpDataReceiverProcessor.create_access_token(
                auth_url=auth_url, credentials=credentials
            )
            headers.update({"Authorization": f"Bearer {auth_token}"})

        # Assumes certs are distributed to the executors beforehand via SparkContext.addFile
        cert_files: Optional[Union[str, Tuple[str, str]]] = None
        if isinstance(cert, tuple):
            cert_files = SparkFiles.get(cert[0]), SparkFiles.get(cert[1])
        elif cert:
            cert_files = SparkFiles.get(cert)

        for row in rows:
            response = HttpDataReceiverProcessor.process_row(
                row=row,
                raise_error=False,  # We don't want to raise error in case of access token expiry
                base_headers=headers,
                cert=cert_files,
                verify=verify,
            )
            if auth_url and response.status_code == status_codes.codes.unauthorized:
                access_token = HttpDataReceiverProcessor.create_access_token(
                    auth_url=auth_url, credentials=credentials
                )
                headers.update({"Authorization": f"Bearer {access_token}"})
                response = HttpDataReceiverProcessor.process_row(
                    row=row,
                    raise_error=raise_error,
                    base_headers=headers,
                    cert=cert_files,
                    verify=verify,
                )
            elif raise_error:
                response.raise_for_status()
            data, is_error = response_processor(response, json.loads(row["state"]))
            if not isinstance(data, list):
                data = [data]
            result.extend(
                [
                    Row(
                        headers=headers,
                        url=row["url"],
                        status=response.status_code,
                        is_error=is_error,
                        error_data=json.dumps(item if is_error else None),
                        success_data=json.dumps(None if is_error else item),
                        state=json.loads(row["state"]),
                    )
                    for item in data
                ]
            )
        return result

    @staticmethod
    def process_row(
        row: Row,
        raise_error: bool,
        base_headers: Dict[str, Any],
        cert: Optional[Union[str, Tuple[str, str]]] = None,
        verify: Optional[Union[bool, str]] = None,
    ) -> Response:
        headers = json.loads(row["headers"])
        headers.update(base_headers)
        http_request = HelixHttpRequest(
            request_type=RequestType.GET,
            url=row["url"],
            headers=headers,
            raise_error=raise_error,
            cert=cert,
            verify=verify,
        )
        return http_request.get_response()



import os
from pathlib import Path
from typing import Dict, Any, Optional, List, Union

from spark_pipeline_framework.utilities.capture_parameters import capture_parameters
from pyspark.ml.param import Param
from pyspark.sql.dataframe import DataFrame
from pyspark.sql.functions import col

from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)


class FrameworkDataFrameAnalyzer(FrameworkTransformer):
    # noinspection PyUnusedLocal
    @capture_parameters
    def __init__(
        self,
        view: str,
        analysis_views_prefix: Optional[str] = None,
        output_folder: Optional[Union[Path, str]] = None,
        columns_to_analyze: Optional[List[str]] = None,
        columns_to_skip: Optional[List[str]] = None,
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
    ):
        """
        For each column in the view, this transformer can either:
        1. Create a view that contains the unique values and the count of each
        2. Write a csv to output_folder that contains the unique values and the count of each
        :param view: view to load the data from
        :param analysis_views_prefix: (Optional) prefix to use when creating the analysis views
        :param output_folder: (Optional) folder in which to create the csvs
        :param columns_to_analyze: (Optional) limit analysis to these columns
        :param columns_to_skip: (Optional) don't include these columns in analysis
        """
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )

        self.logger = get_logger(__name__)

        assert view
        assert output_folder

        self.view: Param[str] = Param(self, "view", "")
        self._setDefault(view=view)

        self.analysis_views_prefix: Param[Optional[str]] = Param(
            self, "analysis_views_prefix", ""
        )
        self._setDefault(analysis_views_prefix=analysis_views_prefix)

        self.output_folder: Param[Optional[Union[Path, str]]] = Param(
            self, "output_folder", ""
        )
        self._setDefault(output_folder=output_folder)

        self.columns_to_analyze: Param[Optional[List[str]]] = Param(
            self, "columns_to_analyze", ""
        )
        self._setDefault(columns_to_analyze=columns_to_analyze)

        self.columns_to_skip: Param[Optional[List[str]]] = Param(
            self, "columns_to_skip", ""
        )
        self._setDefault(columns_to_skip=columns_to_skip)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    def _transform(self, df: DataFrame) -> DataFrame:
        view: str = self.getView()
        analysis_views_prefix: Optional[str] = self.getAnalysisViewsPrefix()
        output_folder: Optional[Union[Path, str]] = self.getOutputFolder()
        columns_to_analyze: Optional[List[str]] = self.getColumnsToAnalyze()
        columns_to_skip: Optional[List[str]] = self.getColumnsToSkip()
        progress_logger: Optional[ProgressLogger] = self.getProgressLogger()

        # get columns in data frame
        df = df.sparkSession.table(view)

        if columns_to_analyze:
            columns_to_analyze = [c for c in df.columns if c in columns_to_analyze]
        else:
            columns_to_analyze = df.columns

        assert columns_to_analyze

        if columns_to_skip:
            columns_to_analyze = [
                c for c in columns_to_analyze if c not in columns_to_skip
            ]

        column_name: str
        for column_name in columns_to_analyze:
            result_df: DataFrame = (
                df.select(column_name)
                .groupBy(column_name)
                .count()
                .orderBy(col("count").desc())
            )
            if output_folder:
                target_path: str = str(os.path.join(str(output_folder), column_name))
                if progress_logger:
                    progress_logger.write_to_log(
                        f"Writing analysis for column {column_name} to {target_path}"
                    )
                result_df.coalesce(1).write.csv(
                    target_path, header=True, mode="overwrite"
                )
            if analysis_views_prefix:
                result_df.createOrReplaceTempView(
                    f"{analysis_views_prefix}{column_name}"
                    if analysis_views_prefix.endswith("_")
                    else f"{analysis_views_prefix}_{column_name}"
                )

        return df

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getView(self) -> str:
        return self.getOrDefault(self.view)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getAnalysisViewsPrefix(self) -> Optional[str]:
        return self.getOrDefault(self.analysis_views_prefix)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getOutputFolder(self) -> Optional[Union[Path, str]]:
        return self.getOrDefault(self.output_folder)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getColumnsToAnalyze(self) -> Optional[List[str]]:
        return self.getOrDefault(self.columns_to_analyze)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getColumnsToSkip(self) -> Optional[List[str]]:
        return self.getOrDefault(self.columns_to_skip)



from typing import Dict, List, Optional, Any

# noinspection PyProtectedMember
from pyspark.sql import SparkSession
from pyspark.sql.types import DataType


def register_udfs(session: SparkSession, udf_infos: List[Dict[str, Any]]) -> None:
    """
    Adds registration for udfs
    :param session: Spark session
    :param udf_infos: UDF context info.
        for example,
        [
            {
                "name": "runCqlLibrary",                                     # UDF name for Spark
                "javaClassName": "com.bwell.services.spark.RunCqlLibrary",   # Fully qualified name of java class
                "returnType": MapType(StringType(), StringType())            # Optional. The return type of the registered Java function.
                                                                             #   The value can be either a :class:`pyspark.sql.types.DataType` object or a DDL-formatted type string.
            }
        ]
    """

    for udf_info in udf_infos:
        udf_name: str = str(udf_info.get("name", None))
        java_class_name: str = str(udf_info.get("javaClassName", None))
        return_type: Optional[DataType | str] = udf_info.get(
            "returnType", None
        )  # Optional

        if udf_name and java_class_name:
            session.udf.registerJavaFunction(
                udf_name,
                java_class_name,
                return_type,
            )



from typing import Any, Dict, Optional

# noinspection PyProtectedMember
from spark_pipeline_framework.utilities.capture_parameters import capture_parameters
from pyspark.sql.dataframe import DataFrame
from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)


class DebugTransformer(FrameworkTransformer):
    # noinspection PyUnusedLocal
    @capture_parameters
    def __init__(
        self,
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
        **kwargs: Optional[Any],
    ):
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )

        self.logger = get_logger(__name__)

        self.kwargs = kwargs

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    def _transform(self, df: DataFrame) -> DataFrame:
        print("kwargs: {}".format(self.kwargs))
        i = df.sparkSession.table("preprocessed")

        i.printSchema()
        i.show()

        return df

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getView(self) -> str:
        return self.getOrDefault(self.view)  # type: ignore



from typing import cast

from pyspark.sql import DataFrame
from pyspark.sql.functions import col
from pyspark.sql.types import StructType, StructField

special_character_substitutions = {
    "=": "_eq",
    ">": "_gt",
    "<": "_lt",
    ">=": "_gte",
    "<=": "_lte",
    "+": "_plus",
    "$": "dol",
    "/": "_",
    "-": "_",
    " ": "_",
    ",": "_",
    ";": "_",
    "{": "_",
    "}": "_",
    ")": "_",
    "(": "_",
}


def replace_special_characters_in_columns_in_data_frame(df: DataFrame) -> DataFrame:
    """
    Replaces the column names with column names that do not include any reserved keywords


    :param df:
    :return: data frame
    """
    for col_name in df.columns:
        new_col_name = replace_special_characters_in_string(col_name)

        if new_col_name != col_name:
            df = df.withColumnRenamed(col_name, new_col_name)

    return df


def replace_special_characters_in_string(col_name: str) -> str:
    new_col_name = col_name
    for key, value in special_character_substitutions.items():
        if key in col_name:
            new_col_name = new_col_name.replace(key, value)

    return new_col_name


def replace_characters_in_nested_fields(
    df: DataFrame,
    column_name: str,
    characters_to_replace: str = " ,;{}()\n\t=",
    replace_with: str = "_",
) -> DataFrame:
    schema: StructType = df.schema

    original_properties_schema: StructField = [
        c for c in schema.fields if c.name == column_name
    ][0]

    original_properties_schema_struct: StructType = cast(
        StructType, original_properties_schema.dataType
    )

    original_properties_schema_fields = original_properties_schema_struct.fields

    has_changes: bool = False
    for field in original_properties_schema_fields:
        new_field_name = replace_characters_in_field_name(
            field_name=field.name,
            characters_to_replace=characters_to_replace,
            replace_with=replace_with,
        )
        if new_field_name != field.name:
            has_changes = True
            field.name = new_field_name

    if has_changes:
        df = df.withColumn(
            column_name, col(column_name).cast(original_properties_schema_struct)
        )

    return df


def replace_characters_in_field_name(
    *,
    field_name: str,
    characters_to_replace: str = " ,;{}()\n\t=",
    replace_with: str = "_",
) -> str:
    result: str = field_name
    for special_char in characters_to_replace:
        result = result.replace(special_char, replace_with)
    return result



import hashlib
from collections import namedtuple
from typing import Dict


class RawAddress:
    """
    The address that needs to get standardized
    """

    _Address = namedtuple(
        "_Address",
        ["address_id", "line1", "line2", "city", "state", "zipcode", "country"],
    )

    def __init__(
        self,
        address_id: str,
        line1: str,
        line2: str = "",
        city: str = "",
        state: str = "",
        zipcode: str = "",
        country: str = "US",
    ):
        self.address = self._Address(
            address_id, line1, line2, city, state, zipcode, country
        )

    def to_dict(self) -> Dict[str, str]:
        return dict(self.address._asdict())

    @classmethod
    def from_dict(cls, address_dict: Dict[str, str]) -> "RawAddress":
        return cls(
            address_id=address_dict.get("address_id") or "",
            line1=address_dict.get("line1") or "",
            line2=address_dict.get("line2") or "",
            city=address_dict.get("city") or "",
            state=address_dict.get("state") or "",
            zipcode=address_dict.get("zipcode") or "",
            country=address_dict.get("country") or "",
        )

    def get_id(self) -> str:
        address_id: str = self.address.address_id
        return address_id

    def set_id(self, address_id: str) -> None:
        self.address = self.address._replace(address_id=address_id)

    def to_str(self) -> str:
        a = self.address
        # todo: make sure the online format is legit
        line2 = " " + a.line2 if a.line2 else ""
        addr: str = f"{a.line1}{line2}, {a.city} {a.state} {a.zipcode} {a.country}"
        return addr.replace("  ", " ").strip(" ,")

    def to_hash(self) -> str:
        # reducing variance by cleaning the address
        a_dict: Dict[str, str] = {
            k: str(v if v else "").strip().lower() for k, v in self.to_dict().items()
        }
        a = self._Address(**a_dict)
        addr: str = f"{a.line1}{a.line2}{a.city}{a.state}{a.zipcode}{a.country}"
        # create hash
        return hashlib.sha1(addr.encode()).hexdigest()

    # noinspection PyMethodMayBeStatic
    def _check_id_unique(self) -> bool:
        # todo check if address_id is unique
        return False



from typing import (
    Protocol,
    List,
    Dict,
    Any,
    Optional,
    AsyncGenerator,
    Union,
)

from spark_pipeline_framework.utilities.async_pandas_udf.v1.async_pandas_batch_function_run_context import (
    AsyncPandasBatchFunctionRunContext,
)

# Information on Python 3.12 Typing syntax: https://realpython.com/python312-typing/
# Mypy information on Python 3.12 generics: https://mypy.readthedocs.io/en/stable/generics.html

type AcceptedParametersType = Dict[str, Any] | object
type AcceptedColumnDataType = Dict[str, Any] | int | float | str | bool


class HandlePandasBatchFunction[
    TParameters: AcceptedParametersType,
    TInputColumnDataType: AcceptedColumnDataType,
    TOutputColumnDataType: AcceptedColumnDataType,
](Protocol):
    """
    This is the definition of the function that is called by the Pandas UDF. This function is called with a batch of
    input values and should return a batch of output values.

    T is the type of the parameters passed to the Pandas UDF.
    TDataType is the type of input data.  It can be dict for struct columns otherwise the type of
    the column
    """

    async def __call__(
        self,
        run_context: AsyncPandasBatchFunctionRunContext,
        input_values: List[TInputColumnDataType],
        parameters: Optional[TParameters],
        additional_parameters: Optional[Dict[str, Any]],
    ) -> AsyncGenerator[TOutputColumnDataType, None]:
        """
        This function is called with a batch of input values and should return a batch of output values.

        :param input_values: input values as a list of dictionaries
        :param parameters: additional parameters passed to the Pandas UDF
        :return: output values as a list of dictionaries
        """
        ...


"""
This is the type alias for when we are operating on a full dataframe not on specific columns
"""


type HandlePandasDataFrameBatchFunction[
    TParameters: AcceptedParametersType
] = HandlePandasBatchFunction[TParameters, Dict[str, Any], Dict[str, Any]]


"""
This is the type alias for when we are operating on a single column for type struct, and we want to output a struct
"""


type HandlePandasStructToStructBatchFunction[
    TParameters: AcceptedParametersType
] = HandlePandasBatchFunction[TParameters, Dict[str, Any], Dict[str, Any]]


"""
This is the type alias for when we are operating on a single column for type scalar, and we want to output a struct
"""


type HandlePandasScalarToStructBatchFunction[
    TParameters: AcceptedParametersType
] = HandlePandasBatchFunction[TParameters, Union[int, float, str, bool], Dict[str, Any]]


"""
This is the type alias for when we are operating on a single column for type scalar, and we want to output a struct
"""


type HandlePandasScalarToScalarBatchFunction[
    TParameters: AcceptedParametersType
] = HandlePandasBatchFunction[
    TParameters, Union[int, float, str, bool], Union[int, float, str, bool]
]

"""
This is the type alias for when we are operating on a single column for type struct, and we want to output a scalar
"""

type HandlePandasStructToScalarBatchFunction[
    TParameters: AcceptedParametersType
] = HandlePandasBatchFunction[TParameters, Dict[str, Any], Union[int, float, str, bool]]



import json
from dataclasses import dataclass
from typing import Any, Dict, List, Optional

from pyspark.sql.types import (
    StructType,
    StructField,
    StringType,
    IntegerType,
)


@dataclass
class ElasticSearchResult:
    """
    Represents the result of an ElasticSearch operation
    """

    url: str
    """  The url of the ElasticSearch server """

    success: int
    """ The number of successful operations """

    failed: int
    """ The number of failed operations """

    payload: List[Dict[str, Any]]
    """ The payload that was sent to the server """

    partition_index: int
    """ The index of the partition """

    error: Optional[str]
    """ The error message if the operation failed """

    def to_dict(self) -> Dict[str, Any]:
        return {
            "url": self.url,
            "success": self.success,
            "failed": self.failed,
            "payload": self.payload,
            "partition_index": self.partition_index,
            "error": self.error,
        }

    def to_dict_flatten_payload(self) -> Dict[str, Any]:
        """
        Returns a dictionary with the payload as a json string

        """
        return {
            "url": self.url,
            "success": self.success,
            "failed": self.failed,
            "partition_index": self.partition_index,
            "payload": json.dumps(self.payload),
            "error": self.error,
        }

    @staticmethod
    def get_schema() -> StructType:
        """
        Returns the schema for the ElasticSearchResult
        """
        return StructType(
            [
                StructField("url", StringType(), True),
                StructField("success", IntegerType(), True),
                StructField("failed", IntegerType(), True),
                StructField("payload", StringType(), True),
                StructField("partition_index", IntegerType(), True),
                StructField("error", StringType(), True),
            ]
        )

    def append(self, other: "ElasticSearchResult") -> None:
        """
        Appends the other ElasticSearchResult to this one

        :param other: The other ElasticSearchResult
        """
        self.success += other.success
        self.failed += other.failed
        self.payload.extend(other.payload)
        if other.error:
            self.error = other.error



import os
import inspect
import importlib.util
from pathlib import Path
from typing import Type, List, TypeVar, Generic

T = TypeVar("T")


class DynamicClassLoader(Generic[T]):
    def __init__(self, base_class: Type[T], folder_path: Path | str):
        self.base_class: Type[T] = base_class
        self.folder_path: Path | str = folder_path

    def find_subclasses(self) -> List[Type[T]]:
        subclasses = []

        # Iterate through files in the specified folder
        for filename in os.listdir(self.folder_path):
            if filename.endswith(".py") and filename != "__init__.py":
                module_name = filename[:-3]
                file_path = os.path.join(self.folder_path, filename)

                # Dynamically import the module
                spec = importlib.util.spec_from_file_location(module_name, file_path)
                assert spec
                module = importlib.util.module_from_spec(spec)
                assert module
                assert spec.loader
                spec.loader.exec_module(module)

                # Inspect classes in the module
                for name, obj in inspect.getmembers(module, inspect.isclass):
                    if issubclass(obj, self.base_class) and obj is not self.base_class:
                        subclasses.append(obj)

        return subclasses



from __future__ import annotations

from typing import (
    List,
    cast,
    Callable,
    Iterator,
    Optional,
)

import pandas as pd
from pyspark.sql import Column
from pyspark.sql.pandas.functions import pandas_udf
from pyspark.sql.types import AtomicType

from spark_pipeline_framework.utilities.async_pandas_udf.v1.async_base_pandas_udf import (
    AsyncBasePandasUDF,
)
from spark_pipeline_framework.utilities.async_pandas_udf.v1.async_pandas_udf_parameters import (
    AsyncPandasUdfParameters,
)
from spark_pipeline_framework.utilities.async_pandas_udf.v1.function_types import (
    HandlePandasScalarToScalarBatchFunction,
    AcceptedParametersType,
)

MyColumnDataType = int | float | str | bool


class AsyncPandasScalarColumnToScalarColumnUDF[TParameters: AcceptedParametersType](
    AsyncBasePandasUDF[
        TParameters,
        pd.Series,  # type:ignore[type-arg]
        pd.Series,  # type:ignore[type-arg]
        MyColumnDataType,
        MyColumnDataType,
    ]
):

    def __init__(
        self,
        *,
        async_func: HandlePandasScalarToScalarBatchFunction[TParameters],
        parameters: Optional[TParameters],
        pandas_udf_parameters: AsyncPandasUdfParameters,
    ) -> None:
        """
        This class wraps an async function in a Pandas UDF for use in Spark.  This class is used
        when the input is a scalar column and the output is a scalar column.


        :param async_func: the async function to run
        :param parameters: the parameters to pass to the async function
        """
        super().__init__(
            async_func=async_func,
            parameters=parameters,
            pandas_udf_parameters=pandas_udf_parameters,
        )

    async def get_input_values_from_chunk(
        self, batch: pd.Series  # type:ignore[type-arg]
    ) -> List[MyColumnDataType]:
        input_values: List[MyColumnDataType] = batch.tolist()
        return input_values

    async def create_output_from_dict(
        self, output_values: List[MyColumnDataType]
    ) -> pd.Series:  # type:ignore[type-arg]
        return pd.Series(output_values)

    def my_apply_process_batch_udf(
        self, batch_iter: Iterator[pd.Series]  # type:ignore[type-arg]
    ) -> Iterator[pd.Series]:  # type:ignore[type-arg]
        # Need this so pandas_udf can use type hints on batch_iter
        return super().apply_process_partition_udf(batch_iter)

    def get_pandas_udf(self, return_type: AtomicType) -> Callable[[Column], Column]:
        """
        Returns a Pandas UDF function that can be used in Spark.

        :param return_type: the return type of the Pandas UDF
        :return: a Pandas UDF function
        """
        return cast(
            Callable[[Column], Column],
            pandas_udf(  # type:ignore[call-overload]
                self.my_apply_process_batch_udf,
                returnType=return_type,
            ),
        )



from abc import abstractmethod
from typing import Dict, Any, TypeVar, Type


from spark_pipeline_framework.utilities.helix_geolocation.v2.structures.standardized_address import (
    StandardizedAddress,
)

T = TypeVar("T", bound="BaseVendorApiResponse")


class BaseVendorApiResponse:
    @abstractmethod
    def to_dict(self) -> Dict[str, Any]:
        pass

    @abstractmethod
    def to_standardized_address(self, *, address_id: str) -> StandardizedAddress:
        pass

    @classmethod
    def from_dict(cls: Type[T], response: Dict[str, Any]) -> T:
        return cls(**response)



import json
import math
from typing import Any, Dict, Optional, Callable, Union, Tuple

from pyspark import RDD, StorageLevel
from pyspark.ml.param import Param
from pyspark.sql.dataframe import DataFrame
from pyspark.sql.functions import col, from_json
from pyspark.sql.types import (
    Row,
    StructType,
    DataType,
    StructField,
    StringType,
    LongType,
    ArrayType,
)

from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_log_metric import (
    ProgressLogMetric,
)
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)
from spark_pipeline_framework.transformers.http_data_sender.v2.http_data_sender_processor import (
    HttpDataSenderProcessor,
)
from spark_pipeline_framework.utilities.api_helper.http_request import (
    SingleJsonResult,
    SingleTextResult,
)
from spark_pipeline_framework.utilities.capture_parameters import capture_parameters
from spark_pipeline_framework.utilities.spark_data_frame_helpers import (
    spark_is_data_frame_empty,
)


class HttpDataSender(FrameworkTransformer):
    # noinspection PyUnusedLocal
    @capture_parameters
    def __init__(
        self,
        *,
        # add your parameters here (be sure to add them to setParams below too)
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
        source_view: str,
        view: Optional[str] = None,
        url: Optional[str] = None,
        auth_url: Optional[str] = None,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
        parse_response_as_json: Optional[bool] = True,
        content_type: str = "application/x-www-form-urlencoded",
        post_as_json_formatted_string: Optional[bool] = None,
        batch_count: Optional[int] = None,
        batch_size: Optional[int] = None,
        cache_storage_level: Optional[StorageLevel] = None,
        payload_generator: Optional[Callable[[Dict[str, Any]], Dict[str, Any]]] = None,
        url_generator: Optional[Callable[[Dict[str, Any]], str]] = None,
        response_processor: Optional[
            Callable[[Dict[str, Any], Union[SingleJsonResult, SingleTextResult]], Any]
        ] = None,
        response_schema: Optional[DataType] = None,
        headers: Optional[Dict[str, Any]] = None,
        cert: Optional[Union[str, Tuple[str, str]]] = None,
        verify: Optional[Union[bool, str]] = None,
    ):
        """
        Sends data to http server (usually REST API)


        :param source_view: read the payload (body) from this view
        :param view: (Optional) view to put the responses in
        :param url: url to call
        :param auth_url: (Optional) url to use to authenticate with client credentials
        :param client_id: (Optional) client id to use to authenticate with client credentials
        :param client_secret: (Optional) client secret to use to authenticate with client credentials
        :param parse_response_as_json: (Optional) whether to parse response as json or not (default = True)
        :param content_type: content_type to use when posting
        :param batch_count: (Optional) number of batches to create
        :param batch_size: (Optional) max number of items in a batch
        :param cache_storage_level: (Optional) how to store the cache:
                                    https://sparkbyexamples.com/spark/spark-dataframe-cache-and-persist-explained/.
        :param payload_generator: Callable which can make the payload based on the `source_view` column
        :param url_generator: Callable which can make the url based on the `source_view` column
        :param response_processor: Callable which processes the response
        :param response_schema: Schema returned by `response_processor`
        :param headers: Any additional headers
        :param cert: certificate or ca bundle file path
        :param verify: controls whether the SSL certificate of the server should be verified when making HTTPS requests.
        """
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )

        self.logger = get_logger(__name__)

        # add a param
        self.source_view: Param[str] = Param(self, "source_view", "")
        self._setDefault(source_view=None)

        self.view: Param[Optional[str]] = Param(self, "view", "")
        self._setDefault(view=None)

        self.url: Param[Optional[str]] = Param(self, "url", "")
        self._setDefault(url=None)

        self.auth_url: Param[Optional[str]] = Param(self, "auth_url", "")
        self._setDefault(auth_url=None)

        self.client_id: Param[Optional[str]] = Param(self, "client_id", "")
        self._setDefault(client_id=None)

        self.client_secret: Param[Optional[str]] = Param(self, "client_secret", "")
        self._setDefault(client_secret=None)

        self.content_type: Param[str] = Param(self, "content_type", "")
        self._setDefault(content_type=None)

        self.batch_count: Param[Optional[int]] = Param(self, "batch_count", "")
        self._setDefault(batch_count=None)

        self.batch_size: Param[Optional[int]] = Param(self, "batch_size", "")
        self._setDefault(batch_size=None)

        self.parse_response_as_json: Param[Optional[bool]] = Param(
            self, "parse_response_as_json", ""
        )
        self._setDefault(parse_response_as_json=None)

        self.post_as_json_formatted_string: Param[Optional[bool]] = Param(
            self, "post_as_json_formatted_string", ""
        )
        self._setDefault(post_as_json_formatted_string=None)

        self.payload_generator: Param[
            Optional[Callable[[Dict[str, Any]], Dict[str, Any]]]
        ] = Param(self, "payload_generator", "")
        self._setDefault(payload_generator=None)

        self.url_generator: Param[Optional[Callable[[Dict[str, Any]], str]]] = Param(
            self, "url_generator", ""
        )
        self._setDefault(url_generator=None)

        self.response_processor: Param[
            Optional[
                Callable[
                    [Dict[str, Any], Union[SingleJsonResult, SingleTextResult]], Any
                ]
            ]
        ] = Param(self, "response_processor", "")
        self._setDefault(response_processor=None)

        self.cache_storage_level: Param[Optional[StorageLevel]] = Param(
            self, "cache_storage_level", ""
        )
        self._setDefault(cache_storage_level=None)

        self.response_schema: Param[Optional[DataType]] = Param(
            self, "response_schema", ""
        )
        self._setDefault(response_schema=response_schema)

        self.headers: Param[Optional[Dict[str, Any]]] = Param(self, "headers", "")
        self._setDefault(headers=headers)

        self.cert: Param[Optional[Union[str, Tuple[str, str]]]] = Param(
            self, "cert", ""
        )
        self._setDefault(cert=cert)

        self.verify: Param[Optional[Union[bool, str]]] = Param(self, "verify", "")
        self._setDefault(verify=verify)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    def _transform(self, df: DataFrame) -> DataFrame:
        name: Optional[str] = self.getName()
        auth_url: Optional[str] = self.getOrDefault(self.auth_url)
        client_id: Optional[str] = self.getOrDefault(self.client_id)
        client_secret: Optional[str] = self.getOrDefault(self.client_secret)
        progress_logger = self.getProgressLogger()
        source_view: str = self.getOrDefault(self.source_view)
        view: Optional[str] = self.getOrDefault(self.view)
        url: Optional[str] = self.getOrDefault(self.url)
        parse_response_as_json: Optional[bool] = self.getOrDefault(
            self.parse_response_as_json
        )
        content_type: str = self.getOrDefault(self.content_type)
        batch_count: Optional[int] = self.getOrDefault(self.batch_count)
        batch_size: Optional[int] = self.getOrDefault(self.batch_size)
        post_as_json_formatted_string: Optional[bool] = self.getOrDefault(
            self.post_as_json_formatted_string
        )
        payload_generator: Optional[Callable[[Dict[str, Any]], Dict[str, Any]]] = (
            self.getOrDefault(self.payload_generator)
        )
        url_generator: Optional[Callable[[Dict[str, Any]], str]] = self.getOrDefault(
            self.url_generator
        )
        response_processor: Optional[
            Callable[[Dict[str, Any], Union[SingleJsonResult, SingleTextResult]], Any]
        ] = self.getOrDefault(self.response_processor)
        cache_storage_level: Optional[StorageLevel] = self.getOrDefault(
            self.cache_storage_level
        )
        response_schema: Optional[DataType] = self.getOrDefault(self.response_schema)
        headers: Optional[Dict[str, Any]] = self.getOrDefault(self.headers)
        cert: Optional[Union[str, Tuple[str, str]]] = self.getOrDefault(self.cert)
        verify: Optional[Union[bool, str]] = self.getOrDefault(self.verify)

        df = df.sparkSession.table(source_view)

        if spark_is_data_frame_empty(df=df):
            return df

        if headers and progress_logger:
            progress_logger.write_to_log(
                f"Using headers: {json.dumps(headers, default=str)}"
            )

        with ProgressLogMetric(
            name=f"{name}_fhir_sender", progress_logger=progress_logger
        ):
            desired_partitions: int
            if batch_count:
                desired_partitions = batch_count
            else:
                row_count: int = df.count()
                desired_partitions = (
                    math.ceil(row_count / batch_size)
                    if batch_size and batch_size > 0
                    else row_count
                )
            self.logger.info(f"Total Batches: {desired_partitions}")

            # ---- Now process all the results ----
            rdd: RDD[Row] = df.repartition(
                desired_partitions
            ).rdd.mapPartitionsWithIndex(
                lambda partition_index, rows: HttpDataSenderProcessor.send_partition_to_server(
                    partition_index=partition_index,
                    rows=rows,
                    url=url,
                    content_type=content_type,
                    headers=headers or dict(),
                    post_as_json_formatted_string=post_as_json_formatted_string,
                    parse_response_as_json=parse_response_as_json,
                    client_id=client_id,
                    auth_url=auth_url,
                    client_secret=client_secret,
                    payload_generator=payload_generator,
                    url_generator=url_generator,
                    response_processor=response_processor,
                    cert=cert,
                    verify=verify,
                )
            )

            rdd = (
                rdd.cache()
                if cache_storage_level is None
                else rdd.persist(storageLevel=cache_storage_level)
            )
            result_df: DataFrame
            if response_schema is not None:
                result_df = self.apply_schema(rdd, response_schema)
            else:
                result_df = rdd.toDF()

            result_df = result_df.where(col("url").isNotNull())
            if view:
                result_df.createOrReplaceTempView(view)

            return result_df

    def apply_schema(self, rdd: RDD[Row], response_schema: DataType) -> DataFrame:
        df: DataFrame = rdd.toDF()
        result_schema = df.schema["result"]
        if result_schema.dataType.typeName() == "string":
            assert isinstance(response_schema, (ArrayType, StructType, str))
            df = df.withColumn(
                "result",
                from_json(df.result, schema=response_schema),
            )
        else:
            df = rdd.toDF(
                schema=StructType(
                    [
                        StructField("url", StringType()),
                        StructField("status", LongType()),
                        StructField("result", response_schema),
                        StructField("headers", StringType()),
                        StructField("request_type", StringType()),
                    ]
                )
            )
        return df



from typing import Optional, Mapping, OrderedDict, Tuple, Union, List, Any, cast

import usaddress
import re

from spark_pipeline_framework.utilities.helix_geolocation.v1.raw_address import (
    RawAddress,
)


class AddressParser:
    @staticmethod
    def safe_tag_address(
        address_line: Optional[str],
        tag_mapping: Optional[Mapping[str, str]] = None,
    ) -> Tuple[OrderedDict[str, Union[List[str], str]], str]:
        """
        Parse an address string into its components using the usaddress parser

        :param address_line: address string
        :param tag_mapping: mapping of address components to their respective tags
        :return: dictionary containing the address components using the tag mapping
        """
        if not address_line:
            tagged_address = OrderedDict[str, Union[List[Any], str]]()
            return tagged_address, "Ambiguous"
        try:
            return cast(
                Tuple[OrderedDict[str, Union[List[str], str]], str],
                usaddress.tag(address_line, tag_mapping=tag_mapping),
            )
        except Exception as e:
            tagged_address = OrderedDict[str, Union[List[Any], str]]()
            return tagged_address, "Ambiguous"

    @staticmethod
    def split_address_with_parser(address: str) -> RawAddress | None:
        """
        Split an address string into its components using the usaddress parser


        :param address: address string
        :return: dictionary containing the address components
        """
        if not address:
            return None
        parsed_address: Optional[OrderedDict[str, Union[List[str], str]]]
        parsed_address_type: Optional[str]
        parsed_address, parsed_address_type = AddressParser.safe_tag_address(
            address_line=address,
            tag_mapping={
                "Recipient": "recipient",
                "AddressNumber": "line1",
                "AddressNumberPrefix": "line1",
                "AddressNumberSuffix": "line1",
                "StreetName": "line1",
                "StreetNamePreDirectional": "line1",
                "StreetNamePreModifier": "line1",
                "StreetNamePreType": "line1",
                "StreetNamePostDirectional": "line1",
                "StreetNamePostModifier": "line1",
                "StreetNamePostType": "line1",
                "CornerOf": "line1",
                "IntersectionSeparator": "line1",
                "LandmarkName": "line1",
                "USPSBoxGroupID": "line1",
                "USPSBoxGroupType": "line1",
                "USPSBoxID": "line1",
                "USPSBoxType": "line1",
                "BuildingName": "line2",
                "OccupancyType": "line2",
                "OccupancyIdentifier": "line2",
                "SubaddressIdentifier": "line2",
                "SubaddressType": "line2",
                "PlaceName": "city",
                "StateName": "state",
                "ZipCode": "postal_code",
            },
        )
        return RawAddress.from_dict(parsed_address.__dict__) if parsed_address else None

    @staticmethod
    def split_address_with_regex(address: str) -> RawAddress | None:
        """
        Splits an address string into its components using regex pattern matching

        :param address: address string
        :return: dictionary containing the address components
        """
        if not address:
            return None

        try:
            # Define regex pattern for matching the address components
            pattern = r"(?P<line1>[^,]+), (?P<city>[^,]+), (?P<state>[A-Z]{2}), (?P<zipcode>\d{5})"
            match = re.match(pattern, address)

            if match:
                components = match.groupdict()
                # Add address line 2 as empty since it is not present in the input address
                components["line2"] = ""
                return RawAddress.from_dict(components)
            else:
                return None
        except Exception as e:
            return None

    @staticmethod
    def split_address(address: str) -> RawAddress | None:
        """
        Splits an address string into its components using usaddress parser and regex pattern matching

        :param address: address string
        :return: dictionary containing the address components
        """
        if not address:
            return None
        parsed_address: RawAddress | None = AddressParser.split_address_with_parser(
            address
        )
        if parsed_address is None:
            parsed_address = AddressParser.split_address_with_regex(address)
        return parsed_address



import os
from pathlib import Path
from shutil import rmtree
from typing import Any

import pytest
from pyspark.sql import SparkSession, DataFrame
from pyspark.sql.types import StructType

from create_spark_session import clean_spark_session
from spark_pipeline_framework.progress_logger.progress_logger import (
    ProgressLogger,
)
from spark_pipeline_framework.progress_logger.test.simple_pipeline import SimplePipeline


WAREHOUSE_CONNECTION_STRING = "jdbc:mysql://root:root_password@warehouse:3306/fhir_rpt/schema?rewriteBatchedStatements=true"


@pytest.fixture(scope="function")
def test_setup() -> None:
    data_dir = Path(__file__).parent
    temp_dir = data_dir.joinpath("temp")
    output_dir = temp_dir.joinpath("output")
    if os.path.isdir(output_dir):
        rmtree(output_dir)


async def test_progress_logger_async(
    spark_session: SparkSession, test_setup: Any
) -> None:
    clean_spark_session(spark_session)
    data_dir: Path = Path(__file__).parent.joinpath("./")
    temp_dir: Path = data_dir.joinpath("temp")

    flights_path: str = f"file://{data_dir.joinpath('flights.csv')}"
    export_path: str = str(temp_dir.joinpath("output").joinpath("flights.json"))

    schema = StructType([])

    spark_session.createDataFrame(spark_session.sparkContext.emptyRDD(), schema)

    spark_session.sql("DROP TABLE IF EXISTS default.flights")

    spark_session.createDataFrame(
        [
            (1, "Qureshi", "Imran"),
            (2, "Vidal", "Michael"),
        ],  # noqa: E231
        ["member_id", "last_name", "first_name"],
    ).createOrReplaceTempView("patients")

    source_df: DataFrame = spark_session.table("patients")

    df = source_df.select("member_id")
    df.createOrReplaceTempView("members")

    # Act
    parameters = {
        "flights_path": flights_path,
        "feature_path": data_dir.joinpath(
            "library/features/carriers_multiple_mappings/v1"
        ),
        "foo": "bar",
        "view2": "my_view_2",
        "export_path": export_path,
    }

    with ProgressLogger() as progress_logger:
        pipeline: SimplePipeline = SimplePipeline(
            parameters=parameters, progress_logger=progress_logger
        )
        transformer = pipeline.fit(df)
        await transformer._transform_async(df)



import pytest

from spark_pipeline_framework.utilities.helix_geolocation.v2.structures.raw_address import (
    RawAddress,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.utilities.address_parser import (
    AddressParser,
)


@pytest.fixture
def sample_address() -> str:
    return "123 Main St, Anytown, CA, 12345"


@pytest.fixture
def sample_address_id() -> str:
    return "1"


@pytest.fixture
def parsed_address() -> RawAddress:
    return RawAddress(
        address_id="1",
        line1="123 Main St",
        line2="",
        city="Anytown",
        state="CA",
        zipcode="12345",
    )


@pytest.mark.asyncio
async def test_safe_tag_address(sample_address: str) -> None:
    print()
    # Test the safe_tag_address method
    tagged_address, address_type = AddressParser.safe_tag_address(sample_address)
    print(tagged_address)
    assert address_type == "Street Address", f"tagged_address: {tagged_address}"
    assert tagged_address["AddressNumber"] == "123", f"tagged_address: {tagged_address}"
    assert tagged_address["StreetName"] == "Main", f"tagged_address: {tagged_address}"
    assert (
        tagged_address["StreetNamePostType"] == "St"
    ), f"tagged_address: {tagged_address}"
    assert tagged_address["PlaceName"] == "Anytown", f"tagged_address: {tagged_address}"
    assert tagged_address["StateName"] == "CA", f"tagged_address: {tagged_address}"
    assert tagged_address["ZipCode"] == "12345", f"tagged_address: {tagged_address}"


@pytest.mark.asyncio
async def test_split_address_with_parser(
    sample_address: str, sample_address_id: str, parsed_address: RawAddress
) -> None:
    # Test the split_address_with_parser method
    result = AddressParser.split_address_with_parser(
        address=sample_address, address_id=sample_address_id
    )
    assert result

    result_dict = result.to_dict()
    parsed_address_dict = parsed_address.to_dict()

    del result_dict["internal_id"]
    del parsed_address_dict["internal_id"]

    assert result_dict == parsed_address_dict


@pytest.mark.asyncio
async def test_split_address_with_regex(
    sample_address: str, sample_address_id: str, parsed_address: RawAddress
) -> None:
    # Test the split_address_with_regex method
    result = AddressParser.split_address_with_regex(
        address=sample_address, address_id=sample_address_id
    )
    assert result

    result_dict = result.to_dict()
    parsed_address_dict = parsed_address.to_dict()

    del result_dict["internal_id"]
    del parsed_address_dict["internal_id"]

    assert result_dict == parsed_address_dict


@pytest.mark.asyncio
async def test_split_address(
    sample_address: str, sample_address_id: str, parsed_address: RawAddress
) -> None:
    # Test the split_address method
    result = AddressParser.split_address(
        address=sample_address, address_id=sample_address_id
    )
    assert result

    result_dict = result.to_dict()
    parsed_address_dict = parsed_address.to_dict()

    del result_dict["internal_id"]
    del parsed_address_dict["internal_id"]

    assert result_dict == parsed_address_dict



from typing import Dict, Any, Optional, List

from pyspark.ml.param import Param
from pyspark.sql.dataframe import DataFrame

from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_partitioner.v2.partition_calculator import (
    PartitionCalculator,
)
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)
from spark_pipeline_framework.utilities.capture_parameters import capture_parameters
from spark_pipeline_framework.utilities.spark_partition_helper.v1.spark_partition_helper import (
    SparkPartitionHelper,
)


class FrameworkPartitioner(FrameworkTransformer):
    # noinspection PyUnusedLocal
    @capture_parameters
    def __init__(
        self,
        *,
        view: str,
        num_partitions: Optional[int] = None,
        partition_size: Optional[int] = None,
        partition_by: Optional[List[str]] = None,
        enable_repartitioning: Optional[bool] = True,
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
        force_partition: Optional[bool] = None,
        enable_coalesce: Optional[bool] = True,
        calculate_automatically: Optional[bool] = False,
        output_row_size: Optional[int] = None,
        input_row_size: Optional[int] = None,
        input_row_count: Optional[int] = None,
        percentage_of_memory_to_use: float = 0.5,  # plan on using only half of the memory
        maximum_number_of_partitions: Optional[int] = None,
        executor_count: Optional[int] = None,
        executor_cores: Optional[int] = None,
        executor_memory: Optional[str] = None,
    ):
        """
        Transformer that partitions a DataFrame if needed based on the parameters provided.
        In general, we are trying to optimize the balance.  We want the make sure we have enough partitions
        as number of executors, so executors are not sitting idle.  If we have 10 executors, then we want
        at least 10 partitions.

        But we are also trying to make sure we have the
        partitions as big as the memory will allow, so we don't have the executors having to process multiple
        partitions for a single task.  If we have 10 executors but 200 partitions, then each executor will
        have to process 20 partitions, which is not efficient.

        Lastly, we want to make sure that the partitions are not too big, so we don't run out of memory available
        on the executors.

        If the calculate_automatically parameter is set to True, then this code will automatically calculate
        the number of partitions taking into account dataframe size, the number of executors
        and memory available to each executor.

        :param view: The name of the view to partition
        :param num_partitions: The number of partitions to create
        :param partition_size: The size of each partition
        :param partition_by: The column to partition by
        :param enable_repartitioning: Whether to enable repartitioning
        :param name: The name of the transformer
        :param parameters: The parameters to capture
        :param progress_logger: The progress logger
        :param force_partition: Whether to force partitioning even if the number of partitions is already correct
        :param enable_coalesce: Whether to enable coalescing if trying to reduce number of partitions.
                                This is more efficient than repartitioning, but it may not result in evenly
                                 distributed partitions
        :param calculate_automatically: Whether to calculate the number of partitions automatically based on the
                                        dataframe size, the number of executors and memory available to each executor.
                                        Dataframe size is estimated based on the first row of the dataframe.
                                        This will not be accurate if the rows are not of similar size.
        :param output_row_size: The expected size of the output row in bytes.  If this is provided,
                                        it is used in calculation of partitions by adding it to input row size.
                                        Otherwise, we assume that the output size will be same as input size.
        :param input_row_size: The size of the input row in bytes.  If this is provided, it is used in calculation of
                                partitions by adding it to expected output row size otherwise it is calculated
                                based on the first row of the dataframe
        :param input_row_count: The number of rows in the input dataframe.  If this is provided, it is used
                                in calculation of partitions otherwise it is calculated based on the dataframe count
        :param percentage_of_memory_to_use: The percentage of memory available to use in each executor.
        :param maximum_number_of_partitions: The maximum number of partitions to create.  If the calculated
                                            number of partitions is more than this, then this value is used.
        :param executor_count: The number of executors to use in calculation of partitions.  If this is not provided,
                                then it is calculated based on the spark.executor.instances config
        :param executor_cores: The number of cores per executor to use in calculation of partitions.  If this is not
                                provided, then it is calculated based on the spark.executor.cores config
        :param executor_memory: The memory available to each executor to use in calculation of partitions.  If this is
                                not provided, then it is calculated based on the spark.executor.memory config
        """
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )

        self.logger = get_logger(__name__)

        assert partition_by is None or (
            isinstance(partition_by, list) and len(partition_by) == 1
        ), "Currently we support partitioning by only a single column"

        # add a param
        self.view: Param[str] = Param(self, "view", "")
        self._setDefault(view=view)

        self.force_partition: Param[Optional[bool]] = Param(self, "force_partition", "")
        self._setDefault(force_partition=None)

        self.num_partitions: Param[Optional[int]] = Param(self, "num_partitions", "")
        self._setDefault(num_partitions=num_partitions)

        self.partition_size: Param[Optional[int]] = Param(self, "partition_size", "")
        self._setDefault(partition_size=partition_size)

        self.partition_by: Param[Optional[List[str]]] = Param(self, "partition_by", "")
        self._setDefault(partition_by=partition_by)

        self.enable_repartitioning: Param[Optional[bool]] = Param(
            self, "enable_repartitioning", ""
        )
        self._setDefault(enable_repartitioning=enable_repartitioning)

        self.enable_coalesce: Param[Optional[bool]] = Param(self, "enable_coalesce", "")
        self._setDefault(enable_coalesce=enable_coalesce)

        self.calculate_automatically: Param[Optional[bool]] = Param(
            self, "calculate_automatically", ""
        )
        self._setDefault(calculate_automatically=calculate_automatically)

        self.output_row_size: Param[Optional[int]] = Param(self, "output_row_size", "")
        self._setDefault(output_row_size=output_row_size)

        self.input_row_size: Param[Optional[int]] = Param(self, "input_row_size", "")
        self._setDefault(input_row_size=input_row_size)

        self.input_row_count: Param[Optional[int]] = Param(self, "input_row_count", "")
        self._setDefault(input_row_count=input_row_count)

        self.percentage_of_memory_to_use: Param[float] = Param(
            self, "percentage_of_memory_to_use", ""
        )
        self._setDefault(percentage_of_memory_to_use=percentage_of_memory_to_use)

        self.maximum_number_of_partitions: Param[Optional[int]] = Param(
            self, "maximum_number_of_partitions", ""
        )
        self._setDefault(maximum_number_of_partitions=maximum_number_of_partitions)

        self.executor_count: Param[Optional[int]] = Param(self, "executor_count", "")
        self._setDefault(executor_count=executor_count)

        self.executor_cores: Param[Optional[int]] = Param(self, "executor_cores", "")
        self._setDefault(executor_cores=executor_cores)

        self.executor_memory: Param[Optional[str]] = Param(self, "executor_memory", "")
        self._setDefault(executor_memory=executor_memory)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    def _transform(self, df: DataFrame) -> DataFrame:
        view: str = self.getOrDefault(self.view)
        force_partition: Optional[bool] = self.getOrDefault(self.force_partition)
        num_partitions: Optional[int] = self.getOrDefault(self.num_partitions)
        enable_coalesce: Optional[bool] = self.getOrDefault(self.enable_coalesce)
        partition_size: Optional[int] = self.getOrDefault(self.partition_size)
        partition_by: Optional[List[str]] = self.getOrDefault(self.partition_by)
        enable_repartitioning: Optional[bool] = self.getOrDefault(
            self.enable_repartitioning
        )
        calculate_automatically: Optional[bool] = self.getOrDefault(
            self.calculate_automatically
        )
        output_row_size: Optional[int] = self.getOrDefault(self.output_row_size)
        input_row_size: Optional[int] = self.getOrDefault(self.input_row_size)
        input_row_count: Optional[int] = self.getOrDefault(self.input_row_count)
        percentage_of_memory_to_use: float = self.getOrDefault(
            self.percentage_of_memory_to_use
        )
        maximum_number_of_partitions: Optional[int] = self.getOrDefault(
            self.maximum_number_of_partitions
        )
        executor_count: Optional[int] = self.getOrDefault(self.executor_count)
        executor_cores: Optional[int] = self.getOrDefault(self.executor_cores)
        executor_memory: Optional[str] = self.getOrDefault(self.executor_memory)

        result_df: DataFrame = df.sparkSession.table(view)
        current_partitions: int = result_df.rdd.getNumPartitions()
        self.logger.info(f"view {view} has {current_partitions} partitions")

        if not enable_repartitioning:
            self.logger.info(
                f"enable_repartitioning parameter is False so partitioning is disabled for view {view}."
            )
            result_df.createOrReplaceTempView(view)
            return result_df

        calculated_partitions: Optional[int] = None
        if calculate_automatically:
            calculated_partitions = PartitionCalculator.calculate_ideal_partitions(
                executor_cores=executor_cores,
                executor_count=executor_count,
                executor_memory=executor_memory,
                input_row_count=input_row_count,
                input_row_size=input_row_size,
                maximum_number_of_partitions=maximum_number_of_partitions,
                output_row_size=output_row_size,
                partition_size=partition_size,
                percentage_of_memory_to_use=percentage_of_memory_to_use,
                df=result_df,
                logger=self.logger,
            )

        desired_partitions: int = (
            SparkPartitionHelper.calculate_desired_partitions(
                df=result_df,
                num_partitions=num_partitions,
                partition_size=partition_size,
            )
            if not calculated_partitions
            else calculated_partitions
        )
        self.logger.info(
            f"Checking to re-partition view {view} to {desired_partitions} partitions"
        )
        # Now repartition the dataframe if the desired number of partitions is different
        # from the current number of partitions
        result_df = SparkPartitionHelper.partition_if_needed(
            df=result_df,
            desired_partitions=desired_partitions,
            enable_repartitioning=enable_repartitioning,
            partition_by_column_name=None if not partition_by else partition_by[0],
            force_partition=force_partition,
            enable_coalesce=enable_coalesce,
        )

        current_partitions = result_df.rdd.getNumPartitions()
        self.logger.info(
            f"view {view} has {current_partitions} partitions after repartition"
        )
        result_df.createOrReplaceTempView(view)

        return result_df



import json
import math
import traceback
from os import environ
from pathlib import Path
from typing import Any, Dict, List, Optional, Union, Callable, Collection

from helix_fhir_client_sdk.structures.get_access_token_result import (
    GetAccessTokenResult,
)
from pyspark import StorageLevel
from pyspark.ml.param import Param
from pyspark.rdd import RDD
from pyspark.sql.dataframe import DataFrame
from pyspark.sql.functions import col, get_json_object
from pyspark.sql.types import Row
from pyspark.errors import AnalysisException

from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_log_metric import (
    ProgressLogMetric,
)
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.utilities.fhir_helpers.fhir_sender_operation import (
    FhirSenderOperation,
)
from spark_pipeline_framework.transformers.fhir_sender.v1.fhir_sender_processor import (
    FhirSenderProcessor,
)
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)

# noinspection PyProtectedMember
from spark_pipeline_framework.utilities.capture_parameters import capture_parameters
from spark_pipeline_framework.utilities.fhir_helpers.fhir_get_access_token import (
    fhir_get_access_token,
)
from spark_pipeline_framework.utilities.fhir_helpers.fhir_merge_response_item_schema import (
    FhirMergeResponseItemSchema,
)

from spark_pipeline_framework.utilities.fhir_helpers.fhir_sender_validation_exception import (
    FhirSenderValidationException,
)
from spark_pipeline_framework.utilities.file_modes import FileWriteModes
from spark_pipeline_framework.utilities.flattener.flattener import flatten
from spark_pipeline_framework.utilities.spark_data_frame_helpers import (
    spark_is_data_frame_empty,
)
from spark_pipeline_framework.utilities.map_functions import remove_field_from_json


class FhirSender(FrameworkTransformer):
    # noinspection PyUnusedLocal
    @capture_parameters
    def __init__(
        self,
        server_url: str,
        file_path: Union[Path, str, Callable[[Optional[str]], Union[Path, str]]],
        response_path: Union[Path, str, Callable[[Optional[str]], Union[Path, str]]],
        resource: str,
        batch_size: Optional[int] = 30,
        limit: int = -1,
        file_format: str = "json",
        validation_server_url: Optional[str] = None,
        throw_exception_on_validation_failure: Optional[bool] = None,
        auth_server_url: Optional[str] = None,
        auth_client_id: Optional[str] = None,
        auth_client_secret: Optional[str] = None,
        auth_login_token: Optional[str] = None,
        auth_scopes: Optional[List[str]] = None,
        auth_well_known_url: Optional[str] = None,
        operation: Union[
            FhirSenderOperation, str
        ] = FhirSenderOperation.FHIR_OPERATION_MERGE.value,
        name: Optional[str] = None,
        additional_request_headers: Optional[Dict[str, str]] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
        mode: str = FileWriteModes.MODE_OVERWRITE,
        error_view: Optional[str] = None,
        view: Optional[str] = None,
        ignore_status_codes: Optional[List[int]] = None,
        retry_count: Optional[int] = None,
        exclude_status_codes_from_retry: Optional[List[int]] = None,
        num_partitions: Optional[int] = None,
        delta_lake_table: Optional[str] = None,
        cache_storage_level: Optional[StorageLevel] = None,
        run_synchronously: Optional[bool] = None,
        sort_by_column_name_and_type: Optional[tuple[str, Any]] = None,
        drop_fields_from_json: Optional[List[str]] = None,
        partition_by_column_name: Optional[str] = None,
        enable_repartitioning: bool = True,
        smart_merge: Optional[bool] = None,
    ):
        """
        Sends FHIR json stored in a folder to a FHIR server

        :param server_url: url for the fhir server
        :param file_path: where to read the json from
        :param response_path: where to store the responses
        :param resource: name of resource to send to fhir server
        :param batch_size: how many resources to send in one batch call to fhir server
        :param validation_server_url: (Optional) url to use for validating FHIR resources
        :param throw_exception_on_validation_failure: whether to continue after a validation failure
        :param auth_server_url: url to auth server
        :param auth_client_id: client id
        :param auth_client_secret: client secret
        :param auth_login_token: login token to send auth server
        :param auth_scopes: scopes to request
        :param additional_request_headers: (Optional) Additional request headers to use
                                            (Eg: {"Accept-Charset": "utf-8"})
        :param operation: What FHIR operation to perform (e.g. $merge, delete, etc.)
        :param mode: if output files exist, should we overwrite or append
        :param error_view: (Optional) log errors into this view (view only exists IF there are errors)
                            and don't throw exceptions.
                            schema: id, resourceType, issue
        :param view: (Optional) store merge result in this view
        :param delta_lake_table: use delta lake format
        :param cache_storage_level: (Optional) how to store the cache:
                                    https://sparkbyexamples.com/spark/spark-dataframe-cache-and-persist-explained/.
        :param run_synchronously: (Optional) Run on the Spark master to make debugging easier on dev machines
        :param sort_by_column_name_and_type: (Optional) tuple of columnName, columnType to be used for sorting
        :param drop_fields_from_json: (Optional) List of field names to drop from json
        :param partition_by_column_name: (Optional) Name of the column that will be used to repartition df
        :param enable_repartitioning: Enable repartitioning or not, default True
        """
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )

        assert (
            isinstance(file_path, Path)
            or isinstance(file_path, str)
            or callable(file_path)
        ), type(file_path)

        assert file_path

        assert progress_logger

        assert server_url

        self.logger = get_logger(__name__)

        self.server_url: Param[str] = Param(self, "server_url", "")
        self._setDefault(server_url=None)

        self.file_path: Param[
            Union[Path, str, Callable[[Optional[str]], Union[Path, str]]]
        ] = Param(self, "file_path", "")
        self._setDefault(file_path=None)

        self.response_path: Param[
            Union[Path, str, Callable[[Optional[str]], Union[Path, str]]]
        ] = Param(self, "response_path", "")
        self._setDefault(response_path=None)

        self.resource: Param[str] = Param(self, "resource", "")
        self._setDefault(resource=None)

        self.limit: Param[int] = Param(self, "limit", "")
        self._setDefault(limit=None)

        self.batch_size: Param[Optional[int]] = Param(self, "batch_size", "")
        self._setDefault(batch_size=batch_size)

        self.throw_exception_on_validation_failure: Param[Optional[bool]] = Param(
            self, "throw_exception_on_validation_failure", ""
        )
        self._setDefault(
            throw_exception_on_validation_failure=throw_exception_on_validation_failure
        )

        self.file_format: Param[str] = Param(self, "file_format", "")
        self._setDefault(file_format=file_format)

        self.validation_server_url: Param[Optional[str]] = Param(
            self, "validation_server_url", ""
        )
        self._setDefault(validation_server_url=None)

        self.auth_server_url: Param[Optional[str]] = Param(self, "auth_server_url", "")
        self._setDefault(auth_server_url=None)

        self.auth_client_id: Param[Optional[str]] = Param(self, "auth_client_id", "")
        self._setDefault(auth_client_id=None)

        self.auth_client_secret: Param[Optional[str]] = Param(
            self, "auth_client_secret", ""
        )
        self._setDefault(auth_client_secret=None)

        self.auth_login_token: Param[Optional[str]] = Param(
            self, "auth_login_token", ""
        )
        self._setDefault(auth_login_token=None)

        self.auth_scopes: Param[List[str]] = Param(self, "auth_scopes", "")
        self._setDefault(auth_scopes=None)

        self.auth_well_known_url: Param[Optional[str]] = Param(
            self, "auth_well_known_url", ""
        )
        self._setDefault(auth_well_known_url=None)

        self.operation: Param[Union[FhirSenderOperation, str]] = Param(
            self, "operation", ""
        )
        self._setDefault(operation=operation)

        self.additional_request_headers: Param[Optional[Dict[str, str]]] = Param(
            self, "additional_request_headers", ""
        )
        self._setDefault(additional_request_headers=additional_request_headers)

        self.mode: Param[str] = Param(self, "mode", "")
        self._setDefault(mode=mode)

        self.error_view: Param[Optional[str]] = Param(self, "error_view", "")
        self._setDefault(error_view=None)

        self.view: Param[Optional[str]] = Param(self, "view", "")
        self._setDefault(view=None)

        self.retry_count: Param[Optional[int]] = Param(self, "retry_count", "")
        self._setDefault(retry_count=None)

        self.exclude_status_codes_from_retry: Param[Optional[List[int]]] = Param(
            self, "exclude_status_codes_from_retry", ""
        )
        self._setDefault(exclude_status_codes_from_retry=None)

        self.num_partitions: Param[Optional[int]] = Param(self, "num_partitions", "")
        self._setDefault(num_partitions=None)

        self.ignore_status_codes: Param[Optional[List[int]]] = Param(
            self, "ignore_status_codes", ""
        )
        self._setDefault(ignore_status_codes=None)

        self.delta_lake_table: Param[Optional[str]] = Param(
            self, "delta_lake_table", ""
        )
        self._setDefault(delta_lake_table=delta_lake_table)

        self.cache_storage_level: Param[Optional[StorageLevel]] = Param(
            self, "cache_storage_level", ""
        )
        self._setDefault(cache_storage_level=None)

        self.run_synchronously: Param[Optional[bool]] = Param(
            self, "run_synchronously", ""
        )
        self._setDefault(run_synchronously=run_synchronously)

        self.sort_by_column_name_and_type: Param[Optional[tuple[str, Any]]] = Param(
            self, "sort_by_column_name_and_type", ""
        )
        self._setDefault(sort_by_column_name_and_type=sort_by_column_name_and_type)

        self.drop_fields_from_json: Param[Optional[List[str]]] = Param(
            self, "drop_fields_from_json", ""
        )
        self._setDefault(drop_fields_from_json=drop_fields_from_json)

        self.partition_by_column_name: Param[Optional[str]] = Param(
            self, "partition_by_column_name", ""
        )
        self._setDefault(partition_by_column_name=partition_by_column_name)

        self.enable_repartitioning: Param[bool] = Param(
            self, "enable_repartitioning", ""
        )
        self._setDefault(enable_repartitioning=enable_repartitioning)

        self.smart_merge: Param[Optional[bool]] = Param(self, "smart_merge", "")

        self._setDefault(smart_merge=None)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    def _transform(self, df: DataFrame) -> DataFrame:
        file_path: Union[Path, str, Callable[[Optional[str]], Union[Path, str]]] = (
            self.getFilePath()
        )
        if callable(file_path):
            file_path = file_path(self.loop_id)
        response_path: Union[Path, str, Callable[[Optional[str]], Union[Path, str]]] = (
            self.getResponsePath()
        )
        if callable(response_path):
            response_path = response_path(self.loop_id)
        name: Optional[str] = self.getName()
        progress_logger: Optional[ProgressLogger] = self.getProgressLogger()
        resource_name: str = self.getResource()
        parameters = self.getParameters()
        additional_request_headers: Optional[Dict[str, str]] = (
            self.getAdditionalRequestHeaders()
        )
        server_url: str = self.getServerUrl()
        batch_size: Optional[int] = self.getBatchSize()
        throw_exception_on_validation_failure: Optional[bool] = (
            self.getThrowExceptionOnValidationFailure()
        )
        operation: Union[FhirSenderOperation, str] = self.getOrDefault(self.operation)
        mode: str = self.getMode()

        delta_lake_table: Optional[str] = self.getOrDefault(self.delta_lake_table)

        cache_storage_level: Optional[StorageLevel] = self.getOrDefault(
            self.cache_storage_level
        )
        sort_by_column_name_and_type: Optional[tuple[str, Any]] = self.getOrDefault(
            self.sort_by_column_name_and_type
        )
        drop_fields_from_json: Optional[List[str]] = self.getOrDefault(
            self.drop_fields_from_json
        )
        partition_by_column_name: Optional[str] = self.getOrDefault(
            self.partition_by_column_name
        )
        smart_merge: Optional[bool] = self.getOrDefault(self.smart_merge)

        if not batch_size or batch_size == 0:
            batch_size = 30
        else:
            batch_size = int(batch_size)

        file_format: str = self.getFileFormat()
        validation_server_url: Optional[str] = self.getValidationServerUrl()
        auth_server_url: Optional[str] = self.getAuthServerUrl()
        auth_well_known_url: Optional[str] = self.getOrDefault(self.auth_well_known_url)
        auth_client_id: Optional[str] = self.getAuthClientId()
        auth_client_secret: Optional[str] = self.getAuthClientSecret()
        auth_login_token: Optional[str] = self.getAuthLoginToken()
        auth_scopes: Optional[List[str]] = self.getAuthScopes()
        auth_access_token: Optional[str] = None

        error_view: Optional[str] = self.getOrDefault(self.error_view)
        view: Optional[str] = self.getOrDefault(self.view)

        log_level: Optional[str] = environ.get("LOGLEVEL")

        ignore_status_codes: List[int] = self.getIgnoreStatusCodes() or []
        ignore_status_codes.append(200)

        retry_count: Optional[int] = self.getOrDefault(self.retry_count)
        exclude_status_codes_from_retry: Optional[List[int]] = self.getOrDefault(
            self.exclude_status_codes_from_retry
        )

        num_partitions: Optional[int] = self.getOrDefault(self.num_partitions)
        enable_repartitioning: bool = self.getOrDefault(self.enable_repartitioning)

        run_synchronously: Optional[bool] = self.getOrDefault(self.run_synchronously)

        if not enable_repartitioning:
            assert (partition_by_column_name or num_partitions) is None

        if parameters and parameters.get("flow_name"):
            if additional_request_headers:
                additional_request_headers.update(
                    {
                        "Origin-Service": f"helix.pipelines:{parameters['flow_name']}",
                    }
                )
            else:
                additional_request_headers = {
                    "Origin-Service": f"helix.pipelines:{parameters['flow_name']}",
                }

        # get access token first so we can reuse it
        if auth_client_id:
            auth_access_token_result: GetAccessTokenResult = fhir_get_access_token(
                logger=self.logger,
                server_url=server_url,
                auth_server_url=auth_server_url,
                auth_client_id=auth_client_id,
                auth_client_secret=auth_client_secret,
                auth_login_token=auth_login_token,
                auth_scopes=auth_scopes,
                log_level=log_level,
                auth_well_known_url=auth_well_known_url,
            )
            auth_access_token = auth_access_token_result.access_token

        self.logger.info(
            f"Calling {server_url}/{resource_name} with client_id={auth_client_id} and scopes={auth_scopes}"
        )
        with ProgressLogMetric(
            name=f"{name}_fhir_sender", progress_logger=progress_logger
        ):
            # read all the files at path into a dataframe
            path_to_files: str = str(file_path)
            try:
                if delta_lake_table:
                    json_df: DataFrame = df.sparkSession.read.format("delta").load(
                        path_to_files
                    )
                elif file_format == "parquet":
                    json_df = df.sparkSession.read.format(file_format).load(
                        path_to_files
                    )
                else:
                    json_df = df.sparkSession.read.text(
                        path_to_files, pathGlobFilter="*.json", recursiveFileLookup=True
                    )

            except AnalysisException as e:
                if str(e).startswith("Path does not exist:"):
                    if progress_logger:
                        progress_logger.write_to_log(f"Folder {path_to_files} is empty")
                    return df
                raise

            row_count: int = json_df.count()
            if not spark_is_data_frame_empty(df=json_df):
                self.logger.info(
                    f"----- Sending {resource_name} (rows={row_count}) to FHIR server {server_url}"
                    f" with batch_size {batch_size}  -----"
                )
                assert batch_size and batch_size > 0
                desired_partitions: int = (
                    (num_partitions or math.ceil(row_count / batch_size))
                    if enable_repartitioning
                    else json_df.rdd.getNumPartitions()
                )
                if partition_by_column_name:
                    json_df = json_df.withColumn(
                        partition_by_column_name,
                        get_json_object(col("value"), f"$.{partition_by_column_name}"),
                    ).repartition(desired_partitions, partition_by_column_name)
                    json_df = json_df.drop(partition_by_column_name)

                if sort_by_column_name_and_type:
                    column_name, column_type = sort_by_column_name_and_type
                    json_df = json_df.withColumn(
                        column_name,
                        get_json_object(col("value"), f"$.{column_name}").cast(
                            column_type
                        ),
                    ).sortWithinPartitions(column_name)
                    json_df = json_df.drop(column_name)

                if drop_fields_from_json:
                    json_schema = json_df.schema
                    # removing sorted field from all the json values
                    json_df = json_df.rdd.map(
                        lambda row: Row(
                            value=remove_field_from_json(
                                row["value"], drop_fields_from_json
                            ),
                        ),
                    ).toDF(json_schema)

                self.logger.info(
                    f"----- Total Batches for {resource_name}: {desired_partitions}  -----"
                )

                result_df: Optional[DataFrame] = None
                if run_synchronously:
                    rows_to_send: List[Row] = json_df.collect()
                    result_rows_list: List[List[Dict[str, Any]]] = list(
                        FhirSenderProcessor.send_partition_to_server(
                            partition_index=0,
                            rows=rows_to_send,
                            desired_partitions=desired_partitions,
                            operation=operation,
                            server_url=server_url,
                            resource_name=resource_name,
                            name=name,
                            auth_server_url=auth_server_url,
                            auth_client_id=auth_client_id,
                            auth_client_secret=auth_client_secret,
                            auth_login_token=auth_login_token,
                            auth_scopes=auth_scopes,
                            auth_access_token=auth_access_token,
                            additional_request_headers=additional_request_headers,
                            log_level=log_level,
                            batch_size=batch_size,
                            validation_server_url=validation_server_url,
                            retry_count=retry_count,
                            exclude_status_codes_from_retry=exclude_status_codes_from_retry,
                            smart_merge=smart_merge,
                        )
                    )
                    result_rows: List[Dict[str, Any]] = flatten(result_rows_list)
                    result_df = df.sparkSession.createDataFrame(  # type: ignore[type-var]
                        result_rows, schema=FhirMergeResponseItemSchema.get_schema()
                    )
                else:
                    # ---- Now process all the results ----
                    if enable_repartitioning and not partition_by_column_name:
                        # repartition only when we haven't already repartitioned by column and
                        # enable_repartitioning is True
                        json_df = json_df.repartition(desired_partitions)
                    rdd: RDD[
                        Union[List[Dict[str, Any]], List[List[Dict[str, Any]]]]
                    ] = json_df.rdd.mapPartitionsWithIndex(
                        lambda partition_index, rows: FhirSenderProcessor.send_partition_to_server(
                            partition_index=partition_index,
                            rows=rows,
                            desired_partitions=desired_partitions,
                            operation=operation,
                            server_url=server_url,
                            resource_name=resource_name,
                            name=name,
                            auth_server_url=auth_server_url,
                            auth_client_id=auth_client_id,
                            auth_client_secret=auth_client_secret,
                            auth_login_token=auth_login_token,
                            auth_scopes=auth_scopes,
                            auth_access_token=auth_access_token,
                            additional_request_headers=additional_request_headers,
                            log_level=log_level,
                            batch_size=batch_size,
                            validation_server_url=validation_server_url,
                            retry_count=retry_count,
                            exclude_status_codes_from_retry=exclude_status_codes_from_retry,
                            smart_merge=smart_merge,
                        )
                    )
                    rdd = (
                        rdd.cache()
                        if cache_storage_level is None
                        else rdd.persist(storageLevel=cache_storage_level)
                    )

                    # turn list of list of string to list of strings
                    rdd_type = Union[Dict[str, Any], List[Dict[str, Any]]]
                    rdd_flat: RDD[rdd_type] = rdd.flatMap(lambda a: a).filter(lambda x: True)  # type: ignore

                    # check if RDD contains a list.  If so, flatMap it
                    rdd_first_row_obj = rdd_flat.take(1)
                    assert isinstance(rdd_first_row_obj, list), type(rdd_first_row_obj)
                    if len(rdd_first_row_obj) > 0:
                        rdd_first_row = rdd_first_row_obj[0]
                        rdd1: RDD[Collection[str]]
                        if isinstance(rdd_first_row, list):
                            rdd1 = rdd_flat.flatMap(lambda a: a).filter(lambda x: True)
                        else:
                            rdd1 = rdd_flat  # type: ignore

                        result_df = rdd1.toDF(
                            schema=FhirMergeResponseItemSchema.get_schema()
                        )

                if result_df is not None:
                    try:
                        self.logger.info(
                            f"Executing requests and writing FHIR {resource_name} responses to disk..."
                        )
                        result_df.write.mode(mode).format(file_format).save(
                            str(response_path)
                        )
                        # result_df.show(truncate=False, n=100)
                        self.logger.info(
                            f"Reading from disk and counting rows for {resource_name}..."
                        )
                        result_df = df.sparkSession.read.format(file_format).load(
                            str(response_path)
                        )
                        file_row_count: int = result_df.count()
                        self.logger.info(
                            f"Wrote {file_row_count} FHIR {resource_name} responses to {response_path}"
                        )
                        if view:
                            result_df.createOrReplaceTempView(view)

                        if FhirMergeResponseItemSchema.issue in result_df.columns:
                            # if there are any errors then raise exception
                            first_error_response: Optional[Row] = result_df.where(
                                col(FhirMergeResponseItemSchema.issue).isNotNull()
                            ).first()
                            if first_error_response is not None:
                                if throw_exception_on_validation_failure:
                                    raise FhirSenderValidationException(
                                        url=validation_server_url or server_url,
                                        json_data=json.dumps(
                                            first_error_response[0],
                                            indent=2,
                                            default=str,
                                        ),
                                    )
                                else:
                                    self.logger.info(
                                        f"------- Failed validations for {resource_name} ---------"
                                    )
                                    failed_df = result_df.where(
                                        col(
                                            FhirMergeResponseItemSchema.issue
                                        ).isNotNull()
                                    )
                                    if error_view:
                                        failed_df.createOrReplaceTempView(error_view)
                                    failed_validations: List[str] = (
                                        failed_df.select(
                                            FhirMergeResponseItemSchema.issue
                                        )
                                        .rdd.flatMap(lambda x: x)
                                        .collect()
                                    )
                                    self.logger.info(failed_validations)
                                    self.logger.info(
                                        f"------- End Failed validations for {resource_name} ---------"
                                    )
                    except Exception as e:
                        exception_traceback = "".join(
                            traceback.TracebackException.from_exception(e).format()
                        )
                        self.logger.exception(
                            f"Exception in FHIR Sender: {str(e)}: {exception_traceback}"
                        )
                        # self.logger.error(f"Response: {result_df.collect()}")
                        raise e

        self.logger.info(
            f"----- Finished sending {resource_name} (rows={row_count}) to FHIR server {server_url}  -----"
        )
        return df

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getServerUrl(self) -> str:
        return self.getOrDefault(self.server_url)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getFilePath(
        self,
    ) -> Union[Path, str, Callable[[Optional[str]], Union[Path, str]]]:
        return self.getOrDefault(self.file_path)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getResponsePath(
        self,
    ) -> Union[Path, str, Callable[[Optional[str]], Union[Path, str]]]:
        return self.getOrDefault(self.response_path)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getLimit(self) -> int:
        return self.getOrDefault(self.limit)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getResource(self) -> str:
        return self.getOrDefault(self.resource)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getBatchSize(self) -> Optional[int]:
        return self.getOrDefault(self.batch_size)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getName(self) -> Optional[str]:
        return self.getOrDefault(self.name) or f" - {self.getResource()}"

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getFileFormat(self) -> str:
        return self.getOrDefault(self.file_format)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getValidationServerUrl(self) -> Optional[str]:
        return self.getOrDefault(self.validation_server_url)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getThrowExceptionOnValidationFailure(self) -> Optional[bool]:
        return self.getOrDefault(self.throw_exception_on_validation_failure)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getAuthServerUrl(self) -> Optional[str]:
        return self.getOrDefault(self.auth_server_url)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getAuthClientId(self) -> Optional[str]:
        return self.getOrDefault(self.auth_client_id)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getAuthClientSecret(self) -> Optional[str]:
        return self.getOrDefault(self.auth_client_secret)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getAuthLoginToken(self) -> Optional[str]:
        return self.getOrDefault(self.auth_login_token)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getAuthScopes(self) -> Optional[List[str]]:
        return self.getOrDefault(self.auth_scopes)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getMode(self) -> str:
        return self.getOrDefault(self.mode)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getIgnoreStatusCodes(self) -> Optional[List[int]]:
        return self.getOrDefault(self.ignore_status_codes)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getAdditionalRequestHeaders(self) -> Optional[Dict[str, str]]:
        return self.getOrDefault(self.additional_request_headers)



import pytest
from typing import List, AsyncGenerator, Dict, Any
from unittest.mock import AsyncMock, MagicMock
from spark_pipeline_framework.utilities.helix_geolocation.v2.cache.cache_result import (
    CacheResult,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.cache.document_db_cache_handler import (
    DocumentDBCacheHandler,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.structures.raw_address import (
    RawAddress,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.structures.vendor_response import (
    VendorResponse,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.vendors.vendor_responses.base_vendor_api_response import (
    BaseVendorApiResponse,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.vendors.vendor_responses.melissa_standardizing_vendor_api_response import (
    MelissaStandardizingVendorApiResponse,
)


@pytest.fixture
def raw_addresses() -> List[RawAddress]:
    # Mocked RawAddress objects with required methods for testing
    raw_address_1 = RawAddress(
        address_id="1",
        line1="123 Main St",
        line2="",
        city="Anytown",
        state="CA",
        zipcode="12345",
    )

    raw_address_2 = RawAddress(
        address_id="2",
        line1="456 Main St",
        line2="",
        city="Anytown",
        state="CA",
        zipcode="12345",
    )

    return [raw_address_1, raw_address_2]


@pytest.fixture
def vendor_responses() -> List[VendorResponse[MelissaStandardizingVendorApiResponse]]:
    # Mocked VendorResponse objects for testing
    vendor_response_1 = VendorResponse(
        vendor_name="melissa",
        response_version="v1",
        api_call_response=MelissaStandardizingVendorApiResponse(
            RecordID="1",
            FormattedAddress="123 Main St",
            Locality="Anytown",
            AdministrativeArea="CA",
            PostalCode="12345",
            CountryISO3166_1_Alpha2="US",
            Latitude="37.7749",
            Longitude="-122.4194",
        ),
        related_raw_address=RawAddress(
            address_id="1",
            line1="123 Main St",
            line2="",
            city="Anytown",
            state="CA",
            zipcode="12345",
        ),
    )

    return [vendor_response_1]


@pytest.mark.asyncio
async def test_check_cache(raw_addresses: List[RawAddress]) -> None:
    # Create a mock collection
    mock_collection = MagicMock()

    # Define the async generator function
    async def mock_cursor_find() -> AsyncGenerator[Dict[str, Any], None]:
        yield {
            "address_hash": raw_addresses[0].to_hash(),
            "vendor_name": "melissa",
            "vendor_std_address": MelissaStandardizingVendorApiResponse(
                RecordID="1",
                FormattedAddress="123 Main St",
                Locality="Anytown",
                AdministrativeArea="CA",
                PostalCode="12345",
                CountryISO3166_1_Alpha2="US",
                Latitude="37.7749",
                Longitude="-122.4194",
            ).to_dict(),
            "response_version": "v1",
        }

    # Mock the find method to return the async generator
    mock_collection.find.return_value = mock_cursor_find()

    handler = DocumentDBCacheHandler(collection=mock_collection)

    result = await handler.check_cache(raw_addresses)

    assert isinstance(result, CacheResult)
    assert len(result.found) == 1  # One address found
    assert len(result.not_found) == 1  # One address not found


@pytest.mark.asyncio
async def test_save_to_cache(
    vendor_responses: List[VendorResponse[BaseVendorApiResponse]],
) -> None:
    # Create a mock collection
    mock_collection = MagicMock()

    # Mock bulk_write
    mock_bulk_write = AsyncMock()
    mock_collection.bulk_write = mock_bulk_write

    handler = DocumentDBCacheHandler(collection=mock_collection)

    await handler.save_to_cache(vendor_responses)

    mock_collection.bulk_write.assert_called_once()
    assert (
        mock_bulk_write.bulk_api_result is not None
    )  # Check the bulk API result logging



import json
from typing import Any, Callable, Dict, Generator, List, Optional, Tuple

from requests import Response

from spark_pipeline_framework.utilities.api_helper.http_request import HelixHttpRequest
from spark_pipeline_framework.utilities.capture_parameters import capture_parameters
from pyspark.ml.param import Param
from pyspark.sql.dataframe import DataFrame
from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_log_metric import (
    ProgressLogMetric,
)
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)
from spark_pipeline_framework.utilities.spark_data_frame_helpers import sc


class HttpDataReceiver(FrameworkTransformer):
    """
    This is a generic class to call a http api and return the response
    """

    # noinspection PyUnusedLocal
    @capture_parameters
    def __init__(
        self,
        http_request_generator: Callable[
            [DataFrame, Optional[ProgressLogger]],
            Generator[HelixHttpRequest, None, None],
        ],
        view_name: str,
        name: Optional[str] = None,
        response_processor: Optional[
            Callable[
                [
                    List[Dict[str, Any]],
                    List[Dict[str, Any]],
                    Response,
                    HelixHttpRequest,
                    Optional[ProgressLogger],
                ],
                Tuple[List[Dict[str, Any]], List[Dict[str, Any]]],
            ]
        ] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
        log_response: bool = False,
        error_view: Optional[str] = None,
        raise_error: bool = True,
    ) -> None:
        """
        Transformer to call and receive data from an API


        :param http_request_generator: Generator to build next http request
        :param view_name: name of the view to read the response into
        :param name: name of transformer
        :param parameters: parameters
        :param progress_logger: progress logger
        it supposed to return a HelixHttpRequest to be used to call the API or return None to end the API call loop
        :param response_processor: it can change the result before loading to spark df
        :param error_view: (Optional) log the details of the api failure into `error_view` view.
        :param raise_error: (Optional) Raise error in case of api failure
        """
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )

        self.logger = get_logger(__name__)

        self.http_request_generator: Param[
            Callable[
                [DataFrame, Optional[ProgressLogger]],
                Generator[HelixHttpRequest, None, None],
            ]
        ] = Param(self, "http_request_generator", "")
        self._setDefault(http_request_generator=None)

        self.view_name: Param[str] = Param(self, "view_name", "")
        self._setDefault(view_name=None)

        self.log_response: Param[bool] = Param(self, "log_response", "")
        self._setDefault(log_response=log_response)

        self.response_processor: Param[
            Optional[
                Callable[
                    [
                        List[Dict[str, Any]],
                        List[Dict[str, Any]],
                        Response,
                        HelixHttpRequest,
                        Optional[ProgressLogger],
                    ],
                    Tuple[List[Dict[str, Any]], List[Dict[str, Any]]],
                ]
            ]
        ] = Param(self, "response_processor", "")
        self._setDefault(response_processor=None)

        self.error_view: Param[Optional[str]] = Param(self, "error_view", "")
        self._setDefault(error_view=None)

        self.raise_error: Param[bool] = Param(self, "raise_error", "")
        self._setDefault(raise_error=raise_error)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    def _transform(self, df: DataFrame) -> DataFrame:
        http_request_generator: Callable[
            [DataFrame, Optional[ProgressLogger]],
            Generator[HelixHttpRequest, None, None],
        ] = self.getHttpRequestGenerator()
        raise_error: bool = self.getRaiseError()
        error_view: Optional[str] = self.getOrDefault(self.error_view)
        view_name = self.getView()
        name: Optional[str] = self.getName()

        progress_logger: Optional[ProgressLogger] = self.getProgressLogger()

        with ProgressLogMetric(
            name=f"{name}_http_data_receiver", progress_logger=progress_logger
        ):
            success_responses: List[Dict[str, Any]] = []
            error_responses: List[Dict[str, Any]] = []
            try:
                for http_request in http_request_generator(df, progress_logger):
                    # Added the statement so the error is not immediately raised
                    http_request.set_raise_error(raise_error)
                    self.logger.debug(f"Calling API: {http_request.to_string()}...")
                    if progress_logger:
                        progress_logger.write_to_log(
                            f"Calling API: {http_request.to_string()}..."
                        )
                    response = http_request.get_response()

                    self.logger.info(
                        f"Api processed: {http_request.url} with status {response.status_code}"
                    )
                    if self.getLogResponse():
                        self.logger.info(f"Response: {response.text}")
                    if progress_logger:
                        progress_logger.write_to_log(
                            f"Api processed: {http_request.url} with status {response.status_code}"
                        )
                        if self.getLogResponse():
                            progress_logger.write_to_log(
                                f"Response [{response.status_code}]: {response.text}"
                            )

                    # accumulated responses before loading to spark
                    response_processor = self.getResponseProcessor()

                    if response_processor:
                        success_responses, error_responses = response_processor(
                            success_responses,
                            error_responses,
                            response,
                            http_request,
                            progress_logger,
                        )
                    else:
                        if response.ok:
                            success_responses.append(response.json())
                        else:
                            error_responses.append(
                                {
                                    "request": {
                                        "url": http_request.url,
                                        "headers": http_request.headers,
                                        "method": http_request.request_type,
                                    },
                                    "response": {
                                        "content": response.text,
                                        "status_code": response.status_code,
                                        "url": response.url,
                                        "headers": response.headers,
                                    },
                                }
                            )

                if error_responses and error_view:
                    error_responses_df = df.sparkSession.read.json(
                        sc(df).parallelize([json.dumps(r) for r in error_responses])
                    )
                    error_responses_df.createOrReplaceTempView(error_view)

                df2 = df.sparkSession.read.json(
                    sc(df).parallelize([json.dumps(r) for r in success_responses])
                )
                df2.createOrReplaceTempView(view_name)
            except Exception as e:
                url = (
                    http_request.url
                    if (http_request and hasattr(http_request, "url"))
                    else None
                )
                error_text: str = f"Url: {url}: {http_request} {repr(e)}"
                self.logger.error(error_text)
                raise Exception(error_text) from e

        return df

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getHttpRequestGenerator(
        self,
    ) -> Callable[
        [DataFrame, Optional[ProgressLogger]], Generator[HelixHttpRequest, None, None]
    ]:
        return self.getOrDefault(self.http_request_generator)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getView(self) -> str:
        return self.getOrDefault(self.view_name)

    # noinspection PyPep8Naming
    def getResponseProcessor(
        self,
    ) -> Optional[
        Callable[
            [
                List[Dict[str, Any]],
                List[Dict[str, Any]],
                Response,
                HelixHttpRequest,
                Optional[ProgressLogger],
            ],
            Tuple[List[Dict[str, Any]], List[Dict[str, Any]]],
        ]
    ]:
        return self.getOrDefault(self.response_processor)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getLogResponse(self) -> bool:
        return self.getOrDefault(self.log_response)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getRaiseError(self) -> bool:
        return self.getOrDefault(self.raise_error)



from dataclasses import dataclass
from datetime import datetime


@dataclass
class PartitionContext:
    """
    This class is used to store the partition index and the start time of the partition
    """

    """ The index of the partition"""
    partition_index: int

    """ The start time of the partition"""
    partition_start_time: datetime



from pathlib import Path

from pyspark.sql import SparkSession, DataFrame
from spark_pipeline_framework.transformers.framework_csv_loader.v1.framework_csv_loader import (
    FrameworkCsvLoader,
)
from spark_pipeline_framework.utilities.spark_data_frame_helpers import (
    create_empty_dataframe,
)
from spark_pipeline_framework.transformers.framework_fill_na_transformer.v1.framework_fill_na_transformer import (
    FrameworkFillNaTransformer,
)


def test_framework_fill_na_transformer(spark_session: SparkSession) -> None:
    # create a dataframe with the test data
    data_dir: Path = Path(__file__).parent.joinpath("./")

    df: DataFrame = create_empty_dataframe(spark_session=spark_session)

    view: str = "primary_care_protocol"
    FrameworkCsvLoader(
        view=view,
        file_path=data_dir.joinpath("primary_care_protocol.csv"),
        clean_column_names=False,
    ).transform(df)

    # ensure we have all the rows even the ones we want to drop
    result_df: DataFrame = spark_session.table(view)
    result_df = result_df.withColumn(
        "Minimum Age", result_df["Minimum Age"].cast("float")
    )
    result_df.createOrReplaceTempView(view)
    assert 7 == result_df.count()

    # drop the rows with null NPI or null Last Name
    FrameworkFillNaTransformer(
        view=view, column_mapping={"Minimum Age": 1.0, "Maximum Age": "No Limit"}
    ).transform(df)

    # assert we get only the rows with a populated NPI
    result_df = spark_session.table(view)
    assert 7 == result_df.count()
    assert "No Limit" == result_df.select("Maximum Age").collect()[1].__getitem__(
        "Maximum Age"
    )
    assert 24.0 == result_df.agg({"Minimum Age": "sum"}).collect()[0][0]



from typing import Optional

from pyspark import SparkConf
from pyspark.sql import DataFrame

from spark_pipeline_framework.utilities.aws.instance_helper.v1.instance_helper import (
    InstanceHelper,
)


class SparkExecutionInformation:
    def __init__(
        self,
        *,
        df: DataFrame,
        executor_count: Optional[int] = None,
        executor_cores: Optional[int] = None,
        executor_memory: Optional[str] = None,
    ) -> None:
        """
        Calculates and returns information about the current Spark execution environment.  This includes the number of
        executors, the number of cores per executor, and the memory available to each executor.


        :param df: The DataFrame to use to calculate the Spark execution information.
        :param executor_count: The number of executors to use.  If not provided, the number of executors will be
                                calculated based on the Spark configuration.
        :param executor_cores: The number of cores per executor to use.  If not provided, the number of cores per
                                executor will be calculated based on the Spark configuration.
        :param executor_memory: The memory available to each executor.  If not provided, the memory available to each
                                executor will be calculated based on the Spark configuration.
        """
        # Calculate the number of executors
        spark_configuration: SparkConf = df.sparkSession.sparkContext.getConf()
        self.spark_executor_instances: Optional[str] = spark_configuration.get(
            "spark.executor.instances"
        )
        self.spark_cluster_target_workers: Optional[str] = spark_configuration.get(
            "spark.databricks.clusterUsageTags.clusterTargetWorkers"
        )
        self.spark_executor_cores: Optional[str] = spark_configuration.get(
            "spark.executor.cores"
        )
        self.spark_executor_memory: Optional[str] = spark_configuration.get(
            "spark.executor.memory"
        )
        spark_cloud_provider: Optional[str] = spark_configuration.get(
            "spark.databricks.cloudProvider"
        )
        self.spark_worker_node_type: Optional[str] = spark_configuration.get(
            "spark.databricks.workerNodeTypeId"
        )
        # if we're in AWS read instance type from spark.databricks.workerNodeTypeId and look up memory
        if not self.spark_executor_memory and spark_cloud_provider == "AWS":
            if self.spark_worker_node_type:
                self.spark_executor_memory = InstanceHelper.get_instance_memory(
                    instance_type=self.spark_worker_node_type
                )

        self.spark_cluster_target_workers_count: Optional[int] = (
            SparkExecutionInformation.safe_str_to_int(self.spark_cluster_target_workers)
            if self.spark_cluster_target_workers
            else None
        )
        if self.spark_cluster_target_workers_count is not None:
            self.spark_cluster_target_workers_count += 1
        # calculate number of executors
        self.current_executor_instances: int | None = (
            executor_count
            or SparkExecutionInformation.safe_str_to_int(self.spark_executor_instances)
            or self.spark_cluster_target_workers_count
            or 1
        )
        # Calculate number of cores per executor
        self.current_executor_cores: int | None = (
            executor_cores
            or SparkExecutionInformation.safe_str_to_int(self.spark_executor_cores)
            or 1
        )
        # Calculate memory available to each executor
        # figure out whether we need to repartition the dataframe
        self.current_executor_memory: int | None = (
            SparkExecutionInformation.parse_memory_string(
                executor_memory
                # figure out whether we need to repartition the dataframe
            )
            or SparkExecutionInformation.parse_memory_string(self.spark_executor_memory)
        )

    @staticmethod
    def safe_str_to_int(s: Optional[str]) -> Optional[int]:
        if not s:
            return None
        try:
            return int(s)
        except (ValueError, TypeError):
            return None  # Or any other default value you prefer

    @staticmethod
    def parse_memory_string(memory_str: Optional[str]) -> Optional[int]:
        if not memory_str:
            return None
        units = {
            "b": 1,
            "k": 1024,
            "m": 1024**2,
            "g": 1024**3,
            "t": 1024**4,
            "p": 1024**5,
        }

        # Extract the numeric part and the unit part
        value = memory_str[:-1]
        unit = memory_str[-1].lower()

        # Convert the value to an integer and multiply by the appropriate unit
        if unit in units:
            return int(value) * units[unit]
        else:
            return None

    @staticmethod
    def convert_bytes_to_human_readable(
        num: Optional[int], suffix: str = "B"
    ) -> Optional[str]:
        if num is None:
            return None
        for unit in ("", "K", "M", "G", "T"):
            if abs(num) < 1024:
                return f"{num:3.1f}{unit}{suffix}"
            num //= 1024
        return f"{num:.1f}Yi{suffix}"



from abc import ABCMeta
from typing import Dict, List

import structlog

from spark_pipeline_framework.utilities.helix_geolocation.v1.raw_address import (
    RawAddress,
)
from spark_pipeline_framework.utilities.helix_geolocation.v1.standardized_address import (
    StandardizedAddress,
)
from spark_pipeline_framework.utilities.helix_geolocation.v1.vendor_response import (
    VendorResponse,
)

logger = structlog.get_logger(__file__)


class StandardizingVendor(metaclass=ABCMeta):
    """
    for now we just save the version attached to the vendor response, in future we need to expand it
    to have different parsers for possible new versions
    """

    def __init__(self, version: str = "1") -> None:
        self._version: str = version

    def standardize(
        self, raw_addresses: List[RawAddress], max_requests: int = 100
    ) -> List[VendorResponse]:
        """
        returns the vendor specific response from the vendor
        """
        return []

    @staticmethod
    def vendor_specific_to_std(
        vendor_specific_addresses: List[VendorResponse],
    ) -> List[StandardizedAddress]:
        """
        each vendor class knows how to convert its response to StdAddress
        """
        return []

    @staticmethod
    def _add_vendor_to_address(
        addresses: List[Dict[str, str]], vendor: str
    ) -> List[Dict[str, str]]:
        return [dict(a, **{"standardize_vendor": vendor}) for a in addresses]

    def _get_request_credentials(self) -> Dict[str, str]:
        return {}

    @staticmethod
    def batch_request_max_size() -> int:
        return 0

    @staticmethod
    def get_record_id(record: Dict[str, str]) -> str:
        return ""

    @staticmethod
    def get_vendor_name() -> str:
        return ""

    def get_version(self) -> str:
        return self._version

    @staticmethod
    def _to_vendor_response(
        vendor_response: List[Dict[str, str]],
        raw_addresses: List[RawAddress],
        vendor_name: str,
        response_version: str,
    ) -> List[VendorResponse]:
        """
        using RecordID, we want to find the corresponding raw address of a vendor response
        """
        # create the map
        id_response_map = {a.get_id(): a for a in raw_addresses}
        # find and assign
        return [
            VendorResponse(
                api_call_response=r,
                related_raw_address=id_response_map[
                    r.get("RecordID") or r.get("address_id") or ""
                ],
                vendor_name=vendor_name,
                response_version=response_version,
            )
            for r in vendor_response
        ]



import json
from typing import Any, Dict, Optional

from furl import furl
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.utilities.api_helper.http_request import (
    HelixHttpRequest,
    RequestType,
)


class FhirMetaDataCache:
    _cache: Dict[str, Optional[Dict[str, Any]]] = {}

    def __init__(self, progress_logger: Optional[ProgressLogger]) -> None:
        self.progress_logger: Optional[ProgressLogger] = progress_logger

    def get_metadata(self, url: str) -> Optional[Dict[str, Any]]:
        if url in FhirMetaDataCache._cache:
            return FhirMetaDataCache._cache.get(url)

        FhirMetaDataCache._cache[url] = self.get_metadata_from_server(server_url=url)
        return FhirMetaDataCache._cache.get(url)

    def get_metadata_from_server(self, server_url: str) -> Optional[Dict[str, Any]]:
        # noinspection PyBroadException
        url = furl(server_url)
        url /= "metadata"
        try:
            http_request = HelixHttpRequest(
                url=url.url,
                request_type=RequestType.GET,
            )
            response = http_request.get_result()
            if self.progress_logger:
                self.progress_logger.write_to_log(
                    f"Received metadata from {url.url}: {json.dumps(response.result, default=str)}"
                )
            return response.result
        except Exception as e:
            if self.progress_logger:
                self.progress_logger.write_to_log(
                    f"ERROR receiving metadata from {url.url}: {e} {json.dumps(e, default=str)}"
                )
            # server does not support /metadata
            return None



from typing import Optional, List, cast, Any, Dict

from pyspark.sql.types import (
    Row,
)


from spark_pipeline_framework.utilities.fhir_helpers.fhir_get_response_schema import (
    FhirGetResponseSchema,
)


class FhirGetResponseReader:
    def __init__(self, row: Row) -> None:
        """
        Reads a Row object and allows property access to it


        :param row: A Spark Row
        """
        self.row: Row = row

    @property
    def sent(self) -> int:
        return cast(int, self.row[FhirGetResponseSchema.sent])

    @property
    def received(self) -> int:
        return cast(int, self.row[FhirGetResponseSchema.received])

    @property
    def responses(self) -> List[str]:
        return cast(List[str], self.row[FhirGetResponseSchema.responses])

    @property
    def first(self) -> Optional[str]:
        return self._safe_get_string(FhirGetResponseSchema.first)

    @property
    def last(self) -> Optional[str]:
        return self._safe_get_string(FhirGetResponseSchema.last)

    @property
    def error_text(self) -> Optional[str]:
        return self._safe_get_string(FhirGetResponseSchema.error_text)

    @property
    def url(self) -> Optional[str]:
        return self._safe_get_string(FhirGetResponseSchema.url)

    @property
    def status_code(self) -> Optional[int]:
        return (
            self.row[FhirGetResponseSchema.status_code]
            if FhirGetResponseSchema.status_code in self.row
            else None
        )

    @property
    def request_id(self) -> Optional[str]:
        return self._safe_get_string(FhirGetResponseSchema.request_id)

    @property
    def access_token(self) -> Optional[str]:
        return self._safe_get_string(FhirGetResponseSchema.access_token)

    @property
    def extra_context_to_return(self) -> Optional[Dict[str, Any]]:
        return (
            self.row[FhirGetResponseSchema.extra_context_to_return]
            if FhirGetResponseSchema.extra_context_to_return in self.row
            else None
        )

    def _safe_get_string(self, property_name: str) -> Optional[str]:
        return self.row[property_name] if property_name in self.row else None



from typing import Dict, Any, Optional

from pyspark.sql import SparkSession
from spark_pipeline_framework.utilities.capture_parameters import capture_parameters
from pyspark.sql.dataframe import DataFrame
from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)
from spark_pipeline_framework.utilities.spark_data_frame_helpers import (
    create_empty_dataframe,
    spark_list_catalog_table_names,
)


class FrameworkCleanSparkSession(FrameworkTransformer):
    # noinspection PyUnusedLocal
    @capture_parameters
    def __init__(
        self,
        # add your parameters here (be sure to add them to setParams below too)
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
    ) -> None:
        """
        Cleans out all the temp tables in memory


        """
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )

        self.logger = get_logger(__name__)

        # add a param
        # self.view: Param[str] = Param(self, "view", "")
        # self._setDefault(view=view)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    def _transform(self, df: DataFrame) -> DataFrame:
        FrameworkCleanSparkSession.clean_spark_session(df.sparkSession)
        df = create_empty_dataframe(df.sparkSession)
        return df

    @staticmethod
    def clean_spark_session(session: SparkSession) -> None:
        """

        :param session:
        :return:
        """
        table_names = spark_list_catalog_table_names(session)

        for table_name in table_names:
            print(f"Dropping table: {table_name}")
            # noinspection SqlDialectInspection,SqlNoDataSourceInspection
            session.sql(f"DROP TABLE IF EXISTS default.{table_name}")
            # noinspection SqlDialectInspection,SqlNoDataSourceInspection
            session.sql(f"DROP VIEW IF EXISTS default.{table_name}")
            # noinspection SqlDialectInspection,SqlNoDataSourceInspection
            session.sql(f"DROP VIEW IF EXISTS {table_name}")

        session.catalog.clearCache()



from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Union

from pyspark.sql.functions import col

# noinspection PyProtectedMember
from spark_pipeline_framework.utilities.capture_parameters import capture_parameters
from pyspark.ml.param import Param
from pyspark.sql.dataframe import DataFrame
from pyspark.sql.types import StructType, DataType
from pyspark.errors import AnalysisException
from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_log_metric import (
    ProgressLogMetric,
)
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)


class FhirReader(FrameworkTransformer):
    # noinspection PyUnusedLocal
    @capture_parameters
    def __init__(
        self,
        file_path: Union[Path, str, Callable[[Optional[str]], Union[Path, str]]],
        view: Optional[str] = None,
        name: Optional[str] = None,
        schema: Optional[Union[StructType, DataType]] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
        bad_records_path: Optional[
            Union[Path, str, Callable[[Optional[str]], Union[Path, str]]]
        ] = None,
        create_empty_view_if_file_path_not_found: Optional[bool] = False,
        limit: int = -1,
        encoding: Optional[str] = None,
        multi_line: bool = False,
        autodiscover_multiline: bool = False,
        filter_by_resource_type: Optional[str] = None,
        delta_lake_table: Optional[str] = None,
    ):
        """
        Reads json files from path and creates columns corresponding to the FHIR schema


        :param file_path: path to the fhir data to read
        :param view: the view to load the data into
        :param name: a name to use when logging information about this transformer
        :param schema: the schema for the fhir data
        :param parameters: the parameter dictionary
        :param progress_logger: the progress logger
        :param bad_records_path: the path to write files that spark cannot parse
        :param create_empty_view_if_file_path_not_found: create an empty view using the provided schema in
                cases where the file_path may not exist.
                *NOTE if this is true and schema is not provided then the df will have no columns!
        :param multi_line: True = file is multi-line, False = file is ndjson
        :param autodiscover_multiline: Whether to read the first line of the file to discover whether it is multiline
        :param filter_by_resource_type: (Optional) read only these resource types from file
        """

        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )

        assert (
            isinstance(file_path, Path)
            or isinstance(file_path, str)
            or callable(file_path)
        ), type(file_path)
        assert file_path

        self.logger = get_logger(__name__)

        self.view: Param[Optional[str]] = Param(self, "view", "")
        self._setDefault(view=view)

        self.file_path: Param[
            Union[Path, str, Callable[[Optional[str]], Union[Path, str]]]
        ] = Param(self, "file_path", "")
        self._setDefault(file_path=file_path)

        self.limit: Param[int] = Param(self, "limit", "")
        self._setDefault(limit=None)

        self.schema: Param[Optional[Union[StructType, DataType]]] = Param(
            self, "schema", ""
        )
        self._setDefault(schema=None)

        self.bad_records_path: Param[
            Optional[Union[Path, str, Callable[[Optional[str]], Union[Path, str]]]]
        ] = Param(self, "bad_records_path", "")
        self._setDefault(bad_records_path=None)

        self.create_empty_view_if_file_path_not_found: Param[Optional[bool]] = Param(
            self, "create_empty_view_if_file_path_not_found", ""
        )
        self._setDefault(
            create_empty_view_if_file_path_not_found=create_empty_view_if_file_path_not_found
        )

        self.encoding: Param[Optional[str]] = Param(self, "encoding", "")
        self._setDefault(encoding=encoding)

        self.multiLine: Param[bool] = Param(self, "multiLine", "")
        self._setDefault(multiLine=multi_line)
        self._set(multiLine=multi_line)

        self.autodiscover_multiline: Param[bool] = Param(
            self, "autodiscover_multiline", ""
        )
        self._setDefault(autodiscover_multiline=autodiscover_multiline)

        self.filter_by_resource_type: Param[Optional[str]] = Param(
            self, "filter_by_resource_type", ""
        )
        self._setDefault(filter_by_resource_type=None)

        self.delta_lake_table: Param[Optional[str]] = Param(
            self, "delta_lake_table", ""
        )
        self._setDefault(delta_lake_table=None)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    def preprocess(self, df: DataFrame, absolute_paths: List[str]) -> None:
        """
        In pre-processing we try to detect whether the file is a normal json or ndjson
        :param df: DataFrame
        :param absolute_paths: list of paths
        """

        if not self.getAutoDiscoverMultiLine():
            return

        assert absolute_paths
        text_df: DataFrame = df.sparkSession.read.text(absolute_paths)
        # read the first line of the file
        first_line: str = text_df.select("value").limit(1).collect()[0][0]
        if (
            first_line is not None
            and first_line.lstrip().startswith("[")
            or first_line.lstrip().rstrip() == "{"
        ):
            self.setMultiLine(True)
        else:
            self.setMultiLine(False)

    def _transform(self, df: DataFrame) -> DataFrame:
        view: Optional[str] = self.getView()
        file_path: Union[Path, str, Callable[[Optional[str]], Union[Path, str]]] = (
            self.getFilePath()
        )
        if callable(file_path):
            file_path = file_path(self.loop_id)
        name: Optional[str] = self.getName()
        schema: Optional[Union[StructType, DataType]] = self.getSchema()
        bad_records_path: Optional[
            Union[Path, str, Callable[[Optional[str]], Union[Path, str]]]
        ] = self.getBadRecordsPath()
        if callable(bad_records_path):
            bad_records_path = bad_records_path(self.loop_id)
        create_empty_view_if_file_path_not_found: Optional[bool] = (
            self.getCreateEmptyViewIfFilePathNotFound()
        )
        encoding: Optional[str] = self.getEncoding()

        if schema is not None:
            assert isinstance(schema, StructType), type(schema)
        progress_logger: Optional[ProgressLogger] = self.getProgressLogger()

        filter_by_resource_type: Optional[str] = self.getFilterByResourceType()
        # limit: int = self.getLimit()
        delta_lake_table: Optional[str] = self.getOrDefault(self.delta_lake_table)

        file_format: str = "delta" if delta_lake_table else "json"

        with ProgressLogMetric(
            name=f"{name or view}_fhir_reader", progress_logger=progress_logger
        ):
            try:
                self.preprocess(df=df, absolute_paths=[str(file_path)])

                # drop malformed records:
                # https://spark.apache.org/docs/2.3.1/api/java/org/apache/spark/sql/DataFrameReader.html#json-scala.collection.Seq-
                # mode (default PERMISSIVE): allows a mode for dealing with corrupt records during parsing.
                # PERMISSIVE : when it meets a corrupted record, puts the malformed string into a field configured
                #   by columnNameOfCorruptRecord, and sets other fields to null. To keep corrupt records,
                #   an user can set a string type field named columnNameOfCorruptRecord in an user-defined schema.
                #   If a schema does not have the field, it drops corrupt records during parsing.
                #   When inferring a schema, it implicitly adds a columnNameOfCorruptRecord field in an
                #   output schema.
                # DROPMALFORMED : ignores the whole corrupted records.
                # FAILFAST : throws an exception when it meets corrupted records.
                # https://docs.databricks.com/spark/latest/spark-sql/handling-bad-records.html
                reader = df.sparkSession.read
                if bad_records_path:
                    reader = reader.option("badRecordsPath", str(bad_records_path))
                if schema and not delta_lake_table:
                    reader = reader.schema(schema)
                if encoding:
                    reader = reader.option("encoding", encoding)
                for k, v in self.getReaderOptions().items():
                    reader = reader.option(k, v)
                if delta_lake_table:
                    df = reader.format(file_format).load(str(file_path))
                else:
                    df = reader.json(str(file_path))

                assert (
                    "_corrupt_record" not in df.columns
                ), f"Found _corrupt_record after reading the file: {file_path}. "

                if "id" in df.columns:
                    df = df.dropDuplicates(["id"])

                if filter_by_resource_type:
                    df = df.where(
                        col("resourceType").eqNullSafe(filter_by_resource_type)
                    )

                if view:
                    df.createOrReplaceTempView(view)

                self.logger.info(
                    f"Read {df.count()} records from json file {file_path}"
                )

            except AnalysisException as e:
                self.logger.exception(f"File read failed from {file_path}")
                if create_empty_view_if_file_path_not_found and view:
                    if schema:
                        df = df.sparkSession.createDataFrame(
                            df.sparkSession.sparkContext.emptyRDD(), schema
                        )
                        df.createOrReplaceTempView(view)
                    else:
                        df = df.sparkSession.createDataFrame(
                            df.sparkSession.sparkContext.emptyRDD()
                        )
                        df.createOrReplaceTempView(view)
                else:
                    raise e
        return df

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getView(self) -> Optional[str]:
        return self.getOrDefault(self.view)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getFilePath(
        self,
    ) -> Union[Path, str, Callable[[Optional[str]], Union[Path, str]]]:
        return self.getOrDefault(self.file_path)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getLimit(self) -> int:
        return self.getOrDefault(self.limit)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getSchema(self) -> Optional[Union[StructType, DataType]]:
        return self.getOrDefault(self.schema)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getBadRecordsPath(
        self,
    ) -> Optional[Union[Path, str, Callable[[Optional[str]], Union[Path, str]]]]:
        return self.getOrDefault(self.bad_records_path)

    # noinspection PyPep8Naming
    def getCreateEmptyViewIfFilePathNotFound(self) -> Optional[bool]:
        return self.getOrDefault(self.create_empty_view_if_file_path_not_found)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getEncoding(self) -> Optional[str]:
        return self.getOrDefault(self.encoding)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getMultiLine(self) -> bool:
        return self.getOrDefault(self.multiLine)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getAutoDiscoverMultiLine(self) -> bool:
        return self.getOrDefault(self.autodiscover_multiline)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def setMultiLine(self, value: bool) -> "FhirReader":
        # noinspection PyUnresolvedReferences
        self._paramMap[self.multiLine] = value
        return self

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getReaderOptions(self) -> Dict[str, Any]:
        options: Dict[str, Any] = {"multiLine": self.getMultiLine()}
        return options

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getFilterByResourceType(self) -> Optional[str]:
        return self.getOrDefault(self.filter_by_resource_type)

    def as_dict(self) -> Dict[str, Any]:
        my_dict: Dict[str, Any] = super().as_dict()
        my_dict["params"] = {
            k: v for k, v in my_dict["params"].items() if k != "schema"
        }  # schema is just too huge
        return my_dict



import json
from os import path, makedirs
from pathlib import Path
from shutil import rmtree

from pyspark.sql import SparkSession, DataFrame
from pyspark.sql.functions import to_json, col

from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_fhir_meta_updater.v2.framework_fhir_meta_updater import (
    FrameworkFhirMetaUpdater,
)


def test_framework_fhir_meta_updater(spark_session: SparkSession) -> None:
    # Arrange
    data_dir: Path = Path(__file__).parent.joinpath("./")

    temp_folder = data_dir.joinpath("./temp")
    if path.isdir(temp_folder):
        rmtree(temp_folder)
    makedirs(temp_folder)

    df: DataFrame = spark_session.read.format("json").load(
        str(data_dir.joinpath("patient.json"))
    )
    df.createOrReplaceTempView("patients")

    # Act
    with ProgressLogger() as progress_logger:
        FrameworkFhirMetaUpdater(
            resource_type="Patient",
            source_view="patients",
            view="fixed_patients",
            source_url="https://p-hi2.digitaledge.cigna.com/PatientAccess/v1-devportal",
            owner_code="cigna",
            access_code="cigna",
            vendor_code="cigna",
            connection_type_code="proa",
            progress_logger=progress_logger,
        ).transform(df)

    # Assert
    result_df = spark_session.table("fixed_patients")

    meta = json.loads(result_df.select(to_json(col("meta"))).collect()[0][0])
    print(meta)
    print(type(meta))

    expected_meta = {
        "source": "https://p-hi2.digitaledge.cigna.com/PatientAccess/v1-devportal/Patient/0000000000000015116",
        "security": [
            {"system": "https://www.icanbwell.com/owner", "code": "cigna"},
            {"system": "https://www.icanbwell.com/access", "code": "cigna"},
            {"system": "https://www.icanbwell.com/vendor", "code": "cigna"},
            {"system": "https://www.icanbwell.com/connectionType", "code": "proa"},
        ],
    }
    assert meta == expected_meta



from pathlib import Path

from pyspark.sql import SparkSession, DataFrame
from pyspark.sql.types import StructField, ArrayType, StructType
from spark_fhir_schemas.r4.complex_types.humanname import HumanNameSchema
from spark_fhir_schemas.r4.resources.bundle import BundleSchema

from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.fhir_bundle_splitter.v1.fhir_bundle_splitter import (
    FhirBundleSplitter,
)
from spark_pipeline_framework.utilities.spark_data_frame_helpers import (
    create_empty_dataframe,
)


def test_fhir_bundle_splitter(spark_session: SparkSession) -> None:
    # Arrange
    data_dir: Path = Path(__file__).parent.joinpath("./")

    df: DataFrame = create_empty_dataframe(spark_session=spark_session)
    bundle_schema = BundleSchema.get_schema()
    assert isinstance(bundle_schema, StructType)
    df = (
        df.sparkSession.read.schema(bundle_schema)
        .option("multiLine", True)
        .json(str(data_dir.joinpath("test.json")))
    )
    df.createOrReplaceTempView("fhir_bundle")

    # Act
    with ProgressLogger() as progress_logger:
        FhirBundleSplitter(
            view="fhir_bundle", progress_logger=progress_logger
        ).transform(df)

    # Assert
    result_df: DataFrame = spark_session.table("fhir_bundle")
    result_df.select("practitioner").printSchema()

    assert "practitioner" in result_df.columns
    # noinspection SpellCheckingInspection
    assert "organization" in result_df.columns
    assert "location" in result_df.columns

    name_schema = result_df.schema["practitioner"].dataType.elementType["name"]  # type: ignore
    assert str(name_schema) == str(
        StructField("name", ArrayType(HumanNameSchema.get_schema()))
    )



from typing import Dict, Any, List

from pyspark.sql import SparkSession, DataFrame

from spark_pipeline_framework.transformers.address_standardization.v1.address_standardization import (
    AddressStandardization,
)
from spark_pipeline_framework.utilities.helix_geolocation.v1.cache.mock_cache_handler import (
    MockCacheHandler,
)
from spark_pipeline_framework.utilities.helix_geolocation.v1.vendors.mock_standardizing_vendor import (
    MockStandardizingVendor,
)
from spark_pipeline_framework.utilities.spark_data_frame_helpers import (
    create_dataframe_from_dictionary,
)


def test_address_standardization_with_caching(spark_session: SparkSession) -> None:
    view_name: str = "foo"
    address_column_mapping: Dict[str, str] = {
        "line1": "address1",
        "line2": "address2",
        "city": "city",
        "state": "state",
        "zipcode": "zip",
    }
    data: List[Dict[str, Any]] = [
        {
            "address1": "547 haight st",
            "address2": "",
            "city": "san francisco",
            "state": "ca",
            "zip": "23434",
        },
        {
            "address1": "548 haight st",
            "address2": "",
            "city": "san francisco",
            "state": "ca",
            "zip": "23434",
        },
        {"address1": None, "address2": None, "city": None, "state": None, "zip": None},
    ]

    df: DataFrame = create_dataframe_from_dictionary(
        data=data, spark_session=spark_session
    )
    df.createOrReplaceTempView(name=view_name)
    AddressStandardization(
        view=view_name,
        address_column_mapping=address_column_mapping,
        standardizing_vendor=MockStandardizingVendor(),
        cache_handler=MockCacheHandler(),
    ).transform(df)
    final_df: DataFrame = df.sql_ctx.table(view_name)
    # assert that we have long and lat columns on the dataframe
    assert 3 == final_df.count()
    assert "latitude" in final_df.columns
    assert "longitude" in final_df.columns



import time
from datetime import datetime, UTC
from os import environ
from typing import Dict, Any, Optional, Union, List, Callable

from helixtelemetry.telemetry.factory.telemetry_factory import TelemetryFactory
from helixtelemetry.telemetry.spans.telemetry_span_creator import TelemetrySpanCreator
from helixtelemetry.telemetry.spans.telemetry_span_wrapper import TelemetrySpanWrapper
from helixtelemetry.telemetry.structures.telemetry_parent import TelemetryParent

from spark_pipeline_framework.mixins.loop_id_mixin import LoopIdMixin
from spark_pipeline_framework.mixins.telemetry_parent_mixin import TelemetryParentMixin

# noinspection PyProtectedMember
from spark_pipeline_framework.utilities.capture_parameters import capture_parameters
from pyspark.ml import Transformer
from pyspark.ml.param import Param
from pyspark.sql.dataframe import DataFrame
from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)
from spark_pipeline_framework.utilities.spark_data_frame_helpers import (
    create_empty_dataframe,
)


class FrameworkLoopTransformer(FrameworkTransformer):
    # noinspection PyUnusedLocal
    @capture_parameters
    def __init__(
        self,
        stages: Union[List[Transformer], Callable[[], List[Transformer]]],
        sleep_interval_in_seconds: int,
        max_time_in_seconds: Optional[int] = None,
        max_number_of_runs: Optional[Union[int, Callable[[], int]]] = None,
        run_until: Optional[datetime] = None,
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
    ):
        """
        Runs the provided stages in a loop


        :param stages: list of transformers or a function that returns a list of transformers
        :param sleep_interval_in_seconds: how long to sleep in between runs
        :param max_time_in_seconds: (Optional) end the loop after running for this many seconds
        :param max_number_of_runs: (Optional) end the loop after running for this many runs
        :param run_until: (Optional) end the loop after running till this datetime
        """
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )

        self.logger = get_logger(__name__)

        self.sleep_interval_in_seconds: int = sleep_interval_in_seconds
        self.max_time_in_seconds: Optional[int] = max_time_in_seconds
        self.max_number_of_runs: Param[Optional[Union[int, Callable[[], int]]]] = Param(
            self, "max_number_of_runs", ""
        )
        self._setDefault(max_number_of_runs=max_number_of_runs)
        self.run_until: Optional[datetime] = run_until

        self.stages: Union[List[Transformer], Callable[[], List[Transformer]]] = stages

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    async def _transform_async(self, df: DataFrame) -> DataFrame:
        progress_logger: Optional[ProgressLogger] = self.getProgressLogger()
        start_time: float = time.time()
        max_number_of_runs: Optional[Union[int, Callable[[], int]]] = (
            self.get_max_number_of_runs()
        )
        if callable(max_number_of_runs):
            max_number_of_runs = max_number_of_runs()
        if progress_logger is not None:
            progress_logger.write_to_log(
                f"---- Starting loop: "
                + f"sleep_interval_in_seconds: {self.sleep_interval_in_seconds}, "
                + f"max_time_in_seconds: {self.max_time_in_seconds}, "
                + f"run_until: {self.run_until}, "
                + f"max_number_of_runs: {max_number_of_runs} "
                + f" ---------"
            )

        current_run_number: int = 0
        while True:
            # noinspection PyCallingNonCallable
            stages: List[Transformer] = (
                self.stages if isinstance(self.stages, list) else self.stages()
            )
            current_run_number += 1
            if progress_logger is not None:
                progress_logger.write_to_log(
                    f"---- Running loop {current_run_number} for {self.getName()} ---------"
                )
                progress_logger.log_event(
                    event_name=f"Started Loop for {self.getName()}",
                    event_text=str(current_run_number),
                )

            telemetry_span_creator: TelemetrySpanCreator = TelemetryFactory(
                telemetry_parent=self.telemetry_parent
                or TelemetryParent.get_null_parent()
            ).create_telemetry_span_creator(log_level=environ.get("LOGLEVEL"))

            stage: Transformer
            for stage in stages:
                if hasattr(stage, "getName"):
                    # noinspection Mypy
                    stage_name = stage.getName()
                    if stage_name is not None:
                        stage_name = f"{stage_name} ({stage.__class__.__name__})"
                    else:
                        stage_name = stage.__class__.__name__
                else:
                    stage_name = stage.__class__.__name__

                telemetry_span: TelemetrySpanWrapper
                async with telemetry_span_creator.create_telemetry_span_async(
                    name=stage_name,
                    attributes={
                        "loop_id": str(current_run_number),
                    },
                    telemetry_parent=self.telemetry_parent,
                ) as telemetry_span:
                    if isinstance(stage, LoopIdMixin):
                        stage.set_loop_id(self.loop_id)
                    if isinstance(stage, TelemetryParentMixin):
                        stage.set_telemetry_parent(
                            telemetry_parent=telemetry_span.create_child_telemetry_parent()
                        )

                    try:
                        # use a new df everytime to avoid keeping data in memory too long
                        df.unpersist(blocking=True)
                        df = create_empty_dataframe(df.sparkSession)
                        if hasattr(stage, "transform_async"):
                            df = await stage.transform_async(df)
                        else:
                            df = stage.transform(df)
                    except Exception as e:
                        if len(e.args) >= 1:
                            # e.args = (e.args[0] + f" in stage {stage_name}") + e.args[1:]
                            e.args = (f"In Stage ({stage_name})", *e.args)
                        raise e

            if progress_logger is not None:
                progress_logger.write_to_log(
                    f"---- Finished loop {current_run_number} for {self.getName()} ---------"
                )
                progress_logger.log_event(
                    event_name=f"Finished Loop for {self.getName()}",
                    event_text=str(current_run_number),
                )
            time_elapsed: float = time.time() - start_time
            if (
                (
                    not self.max_time_in_seconds
                    or time_elapsed < self.max_time_in_seconds
                )
                and (self.run_until is None or datetime.now(UTC) < self.run_until)
                and (not max_number_of_runs or current_run_number < max_number_of_runs)
            ):
                time.sleep(self.sleep_interval_in_seconds)
            else:
                return df

    def as_dict(self) -> Dict[str, Any]:
        return {
            **(super().as_dict()),
            "stages": (
                [s.as_dict() for s in self.stages]  # type: ignore
                if not callable(self.stages)
                else str(self.stages)
            ),
        }

    def get_max_number_of_runs(self) -> Optional[Union[int, Callable[[], int]]]:
        return self.getOrDefault(self.max_number_of_runs)



import json
from typing import Optional, List, Any, Dict

import usaddress
from pydantic import BaseModel

from spark_pipeline_framework.utilities.helix_geolocation.v2.structures.standardized_address import (
    StandardizedAddress,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.vendors.vendor_responses.base_vendor_api_response import (
    BaseVendorApiResponse,
)


class CensusStandardizingVendorAddress(BaseModel):
    address: Optional[str] = None


class CensusStandardizingVendorBenchmark(BaseModel):
    isDefault: Optional[bool] = None
    benchmarkDescription: Optional[str] = None
    id: Optional[str] = None
    benchmarkName: Optional[str] = None


class CensusStandardizingVendorInput(BaseModel):
    address: CensusStandardizingVendorAddress | None = None
    benchmark: CensusStandardizingVendorBenchmark | None = None


class CensusStandardizingVendorTigerLine(BaseModel):
    side: Optional[str] = None
    tigerLineId: Optional[str] = None


class CensusStandardizingVendorCoordinates(BaseModel):
    x: Optional[float] = None
    y: Optional[float] = None


class CensusStandardizingVendorAddressComponents(BaseModel):
    zip: Optional[str] = None
    streetName: Optional[str] = None
    preType: Optional[str] = None
    city: Optional[str] = None
    preDirection: Optional[str] = None
    suffixDirection: Optional[str] = None
    fromAddress: Optional[str] = None
    state: Optional[str] = None
    suffixType: Optional[str] = None
    toAddress: Optional[str] = None
    suffixQualifier: Optional[str] = None
    preQualifier: Optional[str] = None


class CensusStandardizingVendorAddressMatch(BaseModel):
    tigerLine: CensusStandardizingVendorTigerLine | None = None
    coordinates: CensusStandardizingVendorCoordinates | None = None
    addressComponents: CensusStandardizingVendorAddressComponents | None = None
    matchedAddress: Optional[str] = None


class CensusStandardizingVendorApiResponse(BaseModel, BaseVendorApiResponse):
    class Config:
        extra = "ignore"

    input: CensusStandardizingVendorInput | None = None
    addressMatches: Optional[List[CensusStandardizingVendorAddressMatch]] = None
    address_id: str

    def to_dict(self) -> Dict[str, Any]:
        return self.model_dump()

    @classmethod
    def from_standardized_address(
        cls, standardized_address: StandardizedAddress
    ) -> "CensusStandardizingVendorApiResponse":
        return CensusStandardizingVendorApiResponse(
            address_id=standardized_address.get_internal_id(),
            input=CensusStandardizingVendorInput(
                address=CensusStandardizingVendorAddress(
                    address=standardized_address.to_str()
                ),
                benchmark=CensusStandardizingVendorBenchmark(
                    isDefault=True,
                    benchmarkDescription="Public_AR_Current",
                    id="4",
                    benchmarkName="Public_AR_Current",
                ),
            ),
            addressMatches=[
                CensusStandardizingVendorAddressMatch(
                    tigerLine=CensusStandardizingVendorTigerLine(
                        side="L", tigerLineId="107644663"
                    ),
                    coordinates=CensusStandardizingVendorCoordinates(
                        x=(
                            float(standardized_address.longitude)
                            if standardized_address.longitude
                            else None
                        ),
                        y=(
                            float(standardized_address.latitude)
                            if standardized_address.latitude
                            else None
                        ),
                    ),
                    addressComponents=CensusStandardizingVendorAddressComponents(
                        zip=standardized_address.zipcode,
                        streetName=None,
                        preType=None,
                        city=standardized_address.city,
                        preDirection=None,
                        suffixDirection=None,
                        fromAddress=None,
                        state=standardized_address.state,
                        suffixType=None,
                        toAddress=None,
                        suffixQualifier=None,
                        preQualifier=None,
                    ),
                    matchedAddress=standardized_address.to_str(),
                )
            ],
        )

    def to_standardized_address(self, *, address_id: str) -> StandardizedAddress:
        if not self.addressMatches or len(self.addressMatches) == 0:
            return StandardizedAddress.from_address_id(
                address_id=address_id,
                vendor_name="census",
            )

        # Get the address matches
        address_matches: List[CensusStandardizingVendorAddressMatch] | None = (
            self.addressMatches
        )

        if not address_matches or len(address_matches) == 0:
            return StandardizedAddress.from_address_id(
                address_id=address_id,
                vendor_name="census",
            )
        # Get the first address match
        first_address_match: CensusStandardizingVendorAddressMatch = address_matches[0]

        # Get the coordinates
        coordinates: CensusStandardizingVendorCoordinates | None = (
            first_address_match.coordinates
        )

        # Get the x and y coordinates
        longitude: Optional[float] = coordinates.x if coordinates else None
        latitude: Optional[float] = coordinates.y if coordinates else None

        address_components: CensusStandardizingVendorAddressComponents | None = (
            first_address_match.addressComponents
        )

        if not address_components:
            return StandardizedAddress.from_address_id(
                address_id=address_id,
                vendor_name="census",
            )

        # Get the street number
        # street_number: Optional[str] = address_components.get("fromAddress")
        # street_name: Optional[str] = address_components.get("streetName")
        # street_type: Optional[str] = address_components.get("streetSuffix")
        # pre_type: Optional[str] = address_components.get("preType")
        # pre_direction: Optional[str] = address_components.get("preDirection")
        # pre_qualifier: Optional[str] = address_components.get("preQualifier")
        # suffix_direction: Optional[str] = address_components.get("suffixDirection")
        # suffix_type: Optional[str] = address_components.get("suffixType")
        # suffix_qualifier: Optional[str] = address_components.get("suffixQualifier")
        city: Optional[str] = address_components.city
        state: Optional[str] = address_components.state
        postal_code: Optional[str] = address_components.zip

        # Get matched address
        matched_address: Optional[str] = first_address_match.matchedAddress

        # Construct the address line using all components
        address_line: str = ""
        if matched_address:
            try:
                # Parse the address into tagged components
                tagged_address, address_type = usaddress.tag(matched_address)
            except usaddress.RepeatedLabelError as e:
                print(f"Error parsing address: {e}")
                tagged_address = {}

            # Define the components to include in the address line
            components_to_include = [
                "AddressNumber",
                "StreetNamePreDirectional",
                "StreetNamePreType",
                "StreetName",
                "StreetNamePostType",
                "StreetNamePostDirectional",
            ]

            # Extract the components if they exist
            address_line_components = [
                tagged_address[component]
                for component in components_to_include
                if component in tagged_address
            ]

            # Join the components to form the address line
            address_line = " ".join(address_line_components)

        # Create a new address object
        standardized_address: StandardizedAddress = StandardizedAddress(
            address_id=address_id,
            line1=address_line,
            line2=None,
            city=city,
            state=state,
            zipcode=postal_code,
            longitude=str(longitude) if longitude else None,
            latitude=str(latitude) if latitude else None,
            formatted_address=matched_address,
            standardize_vendor="census",
            country="US",
            county=None,
        )

        return standardized_address

    def to_json(self) -> str:
        return json.dumps(self.to_dict())



from typing import List

from spark_pipeline_framework.logger.yarn_logger import get_logger

from spark_pipeline_framework.utilities.helix_geolocation.v2.cache.cache_handler import (
    CacheHandler,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.cache.cache_result import (
    CacheResult,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.structures.raw_address import (
    RawAddress,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.structures.standardized_address import (
    StandardizedAddress,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.standardizing_vendor import (
    StandardizingVendor,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.structures.vendor_response import (
    VendorResponse,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.vendors.vendor_responses.base_vendor_api_response import (
    BaseVendorApiResponse,
)


class AddressStandardizer:
    def __init__(self) -> None:
        self.logger = get_logger(__file__)

    async def standardize_async(
        self,
        raw_addresses: List[RawAddress],
        cache_handler_obj: CacheHandler,
        vendor_obj: StandardizingVendor[BaseVendorApiResponse],
        # Any is needed to make mypy happy since it cannot understand generics that have subclasses
    ) -> List[StandardizedAddress]:
        # check cache if exists
        assert all(
            [r.get_id() is not None for r in raw_addresses]
        ), f"{vendor_obj.get_vendor_name()} requires all addresses to have an id. {[r.to_dict for r in raw_addresses]}"

        cache_lookup_result: CacheResult = await cache_handler_obj.check_cache(
            raw_addresses
        )
        self.logger.info(
            f"cache lookup result -- not found records: {len(cache_lookup_result.not_found)}"
            f" -- found records: {(len(cache_lookup_result.found))}"
            f" using {cache_handler_obj.__class__.__name__}"
        )
        # check api for non-cached (for every 100 records)
        # do the batching here, sending max batch count of each one to the api and then cache the results of each batch
        max_requests = vendor_obj.batch_request_max_size()
        new_std_addresses: List[StandardizedAddress] = []
        for i in range(0, len(cache_lookup_result.not_found), max_requests):
            addresses_not_found_in_cache = cache_lookup_result.not_found[
                i : i + max_requests
            ]

            vendor_responses_batch: List[VendorResponse[BaseVendorApiResponse]] = (
                await vendor_obj.standardize_async(addresses_not_found_in_cache)
            )
            assert len(vendor_responses_batch) == len(addresses_not_found_in_cache), (
                f"{len(vendor_responses_batch)} != {len(cache_lookup_result.not_found[i: i + max_requests])}."
                f"  {type(vendor_obj)}"
            )

            new_std_addresses.extend(
                vendor_obj.vendor_specific_to_std(
                    vendor_specific_addresses=vendor_responses_batch,
                    raw_addresses=raw_addresses,
                )
            )
            # save new address to cache
            await cache_handler_obj.save_to_cache(vendor_responses_batch)

        # combine and return
        assert len(cache_lookup_result.found) + len(new_std_addresses) == len(
            raw_addresses
        ), (
            f"{len(cache_lookup_result.found)} + {len(new_std_addresses)} != {len(raw_addresses)},"
            f" vendor={type(vendor_obj)}, not_found={len(cache_lookup_result.not_found)}"
        )
        return cache_lookup_result.found + new_std_addresses



from typing import Dict, Any, Optional

# noinspection PyProtectedMember
from spark_pipeline_framework.utilities.capture_parameters import capture_parameters
from pyspark.ml.param import Param
from pyspark.sql.dataframe import DataFrame
from pyspark.sql.functions import col
from pyspark.sql.types import StructType, StructField, DataType
from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)


class FrameworkApplySchemaTransformer(FrameworkTransformer):
    # noinspection PyUnusedLocal
    @capture_parameters
    def __init__(
        self,
        *,
        # add your parameters here (be sure to add them to setParams below too)
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
        view: str,
        schema: StructType,
    ):
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )

        self.logger = get_logger(__name__)

        # add a param
        self.view: Param[str] = Param(self, "view", "")
        self._setDefault(view=view)

        self.schema: Param[StructType] = Param(self, "schema", "")
        self._setDefault(schema=schema)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    def _transform(self, df: DataFrame) -> DataFrame:
        view: str = self.getView()
        df = df.sparkSession.table(view)
        schema: StructType = self.getSchema()
        field: StructField
        for field in schema.fields:
            column_name: str = field.name
            column_type: DataType = field.dataType
            df = df.withColumn(column_name, col(column_name).cast(column_type))
        df.createOrReplaceTempView(view)
        return df

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getView(self) -> str:
        return self.getOrDefault(self.view)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getSchema(self) -> StructType:
        return self.getOrDefault(self.schema)
